from __future__ import annotations

import argparse
import logging
import os
from typing import Any

from staticsite.settings import Settings
from staticsite.site import Site
from staticsite.utils import timings

from . import cli

Fail = cli.Fail
Success = cli.Success

log = logging.getLogger("command")

COMMANDS: list[type[Command]] = []


def register(c: type[Command]) -> type[Command]:
    COMMANDS.append(c)
    return c


class Command(cli.Command):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.settings = Settings()
        self.settings.BUILD_COMMAND = self.NAME

    def load_site(self) -> Site:
        # Instantiate site
        site = Site(settings=self.settings)
        with timings("Loaded site in %fs"):
            site.load()

        # If --debug=list was requested
        if self.args.debug == "list":
            for name in sorted(logging.root.manager.loggerDict):
                print(name)
            raise Success()
        return site


# TODO: enable debugging for specific loggers only
#         parser.add_argument("--debug", nargs="?", action="store", const=True, help="verbose output")


class SiteCommand(Command):
    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)

        # Look for extra settings
        settings_files = ["settings.py", ".staticsite.py"]
        if self.args.project:
            if os.path.isfile(self.args.project):
                # If a project file is mentioned, take its directory as default
                # project root
                settings_file = os.path.abspath(self.args.project)
                settings_dir, settings_file = os.path.split(settings_file)
                if not self.args.project.endswith(".py"):
                    log.warning(
                        "%s: project settings does not end in `.py`: contents ignored",
                        settings_file,
                    )
                else:
                    settings_files = [settings_file]
            else:
                # If a project directory is mentioned, take it as default
                # project root
                settings_dir = os.path.abspath(self.args.project)
        else:
            settings_dir = os.getcwd()

        # Load the first settings file found (if any)
        for relpath in settings_files:
            abspath = os.path.join(settings_dir, relpath)
            if os.path.isfile(abspath):
                log.info("%s: loading settings", abspath)
                self.settings.load(abspath)
                break

        # Set default project root if undefined
        if self.settings.PROJECT_ROOT is None:
            self.settings.PROJECT_ROOT = settings_dir

        # Command line overrides for settings
        if self.args.theme:
            self.settings.THEME = (os.path.abspath(self.args.theme),)
        if self.args.content:
            self.settings.CONTENT = os.path.abspath(self.args.content)
        if self.args.archetypes:
            self.settings.ARCHETYPES = os.path.abspath(self.args.archetypes)
        if self.args.output:
            self.settings.OUTPUT = os.path.abspath(self.args.output)
        if self.args.draft:
            self.settings.DRAFT_MODE = True

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)

        parser.add_argument(
            "project",
            nargs="?",
            help="project directory or .py configuration file (default: the current directory)",
        )
        parser.add_argument(
            "--theme", help="theme directory location. Overrides settings.THEME"
        )
        parser.add_argument(
            "--content", help="content directory location. Overrides settings.CONTENT"
        )
        parser.add_argument(
            "--archetypes",
            help="archetypes directory location. Override settings.ARCHETYPES",
        )
        parser.add_argument(
            "-o", "--output", help="output directory location. Override settings.OUTPUT"
        )
        parser.add_argument(
            "--draft",
            action="store_true",
            help="do not ignore pages with date in the future",
        )

        return parser



from __future__ import annotations

import json
import logging
from collections import defaultdict
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

import jinja2

from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.features.data import DataPage, DataPages
from staticsite.features.links.data import Link, LinkCollection
from staticsite.features.links.index import LinkIndexPage
from staticsite.page import Page, TemplatePage

if TYPE_CHECKING:
    import argparse

    from staticsite import file, fstree
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("links")


class LinksField(TrackedField["LinksPageMixin", list[dict[str, Any]]]):
    """
    Extra metadata for external links.

    It is a list of dicts of metadata, one for each link. In each dict, these keys are recognised:

    * `title`: str: short title for the link
    * `url`: str: external URL
    * `abstract`: str: long description or abstract for the link
    * `archive`: str: URL to an archived version of the site
    * `tags`: List[str]: tags for this link
    * `related`: List[Dict[str, str]]: other related links, as a list of dicts with
      `title` and `url` keys
    """

    tracked_by = "links"

    def _clean(self, page: LinksPageMixin, value: Any) -> list[dict[str, Any]]:
        if not isinstance(value, list):
            raise TypeError(
                f"value for {page!r}.{self.name} has type {type(value).__name__} instead of dict"
            )
        return value


class LinksPageMixin(Page):
    # Allow every page to contribute links information
    links = LinksField()


class LinksTemplatePageMixin(LinksPageMixin, TemplatePage):
    @jinja2.pass_context
    def html_body(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        rendered = super().html_body(context, **kw)

        # If the page has rendered pointers to external links, add a dataset
        # with extra information for them
        if not (external_links := getattr(self, "rendered_external_links", ())):
            return rendered

        feature = cast(Links, self.site.features["links"])
        links = feature.links
        if not feature.indices:
            return rendered

        tag_indices = feature.indices[0].by_tag
        data: dict[str, dict[str, Any]] = {}
        for url in external_links:
            if (linfo := links.get(url)) is None:
                continue
            info = linfo.as_dict()

            # Resolve tag urls for page into a { title: …,  url: … } dict
            tags = info.get("tags")
            if tags:
                tag_dicts = []
                for tag in tags:
                    dest = tag_indices[tag]
                    tag_dicts.append({"tag": tag, "url": self.url_for(dest)})
                info["tags"] = tag_dicts

            data[url] = info
        if data:
            rendered += (
                "\n<script type='application/json' class='links-metadata'>"
                f"{json.dumps(data)}"
                "</script>\n"
            )

        return rendered


class LinksPage(LinksPageMixin, DataPage):
    """
    Page with a link collection posted as metadata only.
    """

    TYPE = "links"

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        if self.links:
            self.link_collection = LinkCollection(
                {link["url"]: Link(link, page=self) for link in self.links}
            )
        else:
            self.link_collection = LinkCollection({})


class Links(PageTrackingMixin[LinksPageMixin], Feature):
    """
    Collect links and link metadata from page metadata.

    ## Annotated external links

    The Links feature allows to annotate external links with extra metadata, like
    an abstract, tags, a URL with an archived version of the site, or other related
    links.

    You can add a `links` list to the metadata of the page, containing a dict for
    each external link to annotate in the page.

    See [metadata documentation](metadata.md) for a reference on the `links` field.

    You can create a [data page](data.md) with a `links` list and a `data-type:
    links` to have the data page render as a collection of links.


    ### Links metadata

    This is the full list of supported links metadata:

    * `url: str`: the link URL
    * `archive: str`: URL to an archived version of the site
    * `title: str`: short title for the link
    * `abstract: str`: long description or abstract for the link
    * `tags: List[str]`: tags for this link
    * `related: List[Dict[str, str]]`: other related links, as a list of dicts with
      `title` and `url` keys


    ## Templates

    The template used to render link collections is `data-links.html`, which works
    both on [data-only](data.md) link collections, and on `.links`-generated pages.


    ## Link indices

    If you add a `name.links` file, empty or containing some metadata, it will be
    rendered as a hierarchy of index pages one for each link tag found.

    `data-links.html` is used as default template for `.links`-generated pages.
    """

    RUN_AFTER = ["data"]

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        self.j2_globals["links_tag_index_url"] = self.links_tag_index_url

        # Shortcut to access the Data feature
        self.data = cast(DataPages, self.site.features["data"])

        # Let the Data feature render link collections from pure yaml files,
        # when they have a `data_type: links` metadata
        self.data.register_page_class("links", LinksPage)

        # Index of all links
        self.links = LinkCollection()

        # Pages for .links files found in the site
        self.indices: list[LinkIndexPage] = []

        self.by_tag: dict[str, LinkCollection]

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        if issubclass(page_cls, TemplatePage):
            return (LinksTemplatePageMixin,)
        else:
            return (LinksPageMixin,)

    def get_used_page_types(self) -> list[type[Page]]:
        return [LinksPage, LinkIndexPage]

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        """
        Handle .links pages that generate the browseable archive of annotated
        external links
        """
        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            if not fname.endswith(".links"):
                continue
            taken.append(fname)

            name = fname[:-6]

            try:
                fm_meta = self.load_file_meta(directory, fname)
            except Exception:
                log.exception("%s: cannot parse taxonomy information", src.relpath)
                continue
            kwargs.update(fm_meta)

            page = node.create_source_page_as_path(
                page_cls=LinkIndexPage,
                src=src,
                name=name,
                links=self,
                link_collection=self.links,
                **kwargs,
            )
            if page is not None:
                pages.append(page)
                self.indices.append(page)

        for fname in taken:
            del files[fname]

        return pages

    def load_file_meta(self, directory: fstree.Tree, fname: str) -> dict[str, Any]:
        """
        Parse the links file to read its description
        """
        from staticsite.utils import front_matter

        with directory.open(fname, "rt") as fd:
            fmt, meta = front_matter.read_whole(fd)
        return meta

    @jinja2.pass_context
    def links_tag_index_url(self, context: jinja2.runtime.Context, tag: str) -> str:
        dest = self.indices[0].by_tag[tag]
        page: Page = context.parent["page"]
        return page.url_for(dest)

    def organize(self) -> None:
        # Index links by tag
        self.by_tag = defaultdict(LinkCollection)
        for page in self.tracked_pages:
            if not page.links:
                continue
            for link_dict in page.links:
                link = Link(link_dict)
                self.links.append(link)
                for tag in link.tags:
                    self.by_tag[tag].append(link)

    def generate(self) -> None:
        # Call analyze on all .links pages, to populate them
        for index in self.indices:
            index.organize()

    def add_site_commands(self, subparsers: argparse._SubParsersAction[Any]) -> None:
        super().add_site_commands(subparsers)
        from staticsite.features.links.cmdline import LinkLint

        LinkLint.add_subparser(subparsers)


FEATURES = {
    "links": Links,
}



import os
from unittest import TestCase

from staticsite.theme import Theme

from . import utils as test_utils


class MockPage:
    def __init__(self, **kw):
        self.meta = kw
        for k, v in kw.items():
            setattr(self, k, v)

    def __str__(self):
        return str(self.site_path)

    def __repr__(self):
        return str(self.site_path)


class TestTemplates(test_utils.MockSiteTestMixin, TestCase):
    def test_arrange(self):
        with self.site({}) as mocksite:
            theme = Theme(mocksite.site, "test", [{"root": ".", "name": "test"}])
            theme.load()

            self.maxDiff = None

            # Small list
            pages = [MockPage(site_path=x, date=-x) for x in range(10)]

            expr = theme.jinja2.compile_expression("pages|arrange('url')")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url')")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True),
            )

            expr = theme.jinja2.compile_expression("pages|arrange('url', 5)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)[:5]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url', 5)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True)[:5],
            )

            expr = theme.jinja2.compile_expression("pages|arrange('url', 15)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url', 15)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True),
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-date')")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.date, reverse=True)
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-date', 5)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.date, reverse=True)[:5]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('date', 5)")
            self.assertEqual(expr(pages=pages), sorted(pages, key=lambda x: x.date)[:5])

            # Limit = 1

            expr = theme.jinja2.compile_expression("pages|arrange('url', 1)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)[:1]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url', 1)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True)[:1],
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-date', 1)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.date, reverse=True)[:1]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('date', 1)")
            self.assertEqual(expr(pages=pages), sorted(pages, key=lambda x: x.date)[:1])

            # Large list

            pages = [MockPage(site_path=x, date=-x) for x in range(100)]

            expr = theme.jinja2.compile_expression("pages|arrange('url', 5)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)[:5]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url', 5)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True)[:5],
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-date', 5)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.date, reverse=True)[:5]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('date', 5)")
            self.assertEqual(expr(pages=pages), sorted(pages, key=lambda x: x.date)[:5])

            expr = theme.jinja2.compile_expression("pages|arrange('url', 50)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.site_path)[:50]
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-url', 50)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.site_path, reverse=True)[:50],
            )

            expr = theme.jinja2.compile_expression("pages|arrange('-date', 50)")
            self.assertEqual(
                expr(pages=pages),
                sorted(pages, key=lambda x: x.date, reverse=True)[:50],
            )

            expr = theme.jinja2.compile_expression("pages|arrange('date', 50)")
            self.assertEqual(
                expr(pages=pages), sorted(pages, key=lambda x: x.date)[:50]
            )

    def test_page_template_conflicts(self):
        self.maxDiff = None

        files = {
            "index.md": {},
            "page.html": "",
        }

        with self.site(files) as mocksite:
            site = mocksite.site
            mocksite.assertPagePaths(("", "page.html"))
            index, page = mocksite.page("", "page.html")

            self.assertEqual(
                index.to_dict(),
                {
                    "src": {
                        "relpath": "index.md",
                        "abspath": os.path.join(site.content_root, "index.md"),
                    },
                    "site_path": "",
                    "build_path": "index.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2019 Test User",
                        "date": "2019-06-01 12:30:00+02:00",
                        "draft": False,
                        "indexed": True,
                        "syndicated": True,
                        "syndication_date": "2019-06-01 12:30:00+02:00",
                        "site_name": "Test site",
                        "site_url": "https://www.example.org",
                        "template": "page.html",
                        "template_copyright": "compiled:None",
                        "title": "Test site",
                        "related": {},
                        "nav": [],
                    },
                    "type": "markdown",
                },
            )

            rendered = index.render().content()
            self.assertIn(b"Test site", rendered)

            self.assertEqual(
                page.to_dict(),
                {
                    "src": {
                        "relpath": "page.html",
                        "abspath": os.path.join(site.content_root, "page.html"),
                    },
                    "site_path": "page.html",
                    "build_path": "page.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2019 Test User",
                        "date": "2019-06-01 12:30:00+02:00",
                        "draft": False,
                        "indexed": True,
                        "syndicated": True,
                        "syndication_date": "2019-06-01 12:30:00+02:00",
                        "site_name": "Test site",
                        "site_url": "https://www.example.org",
                        "template": "compiled:page.html",
                        "template_copyright": "compiled:None",
                        "title": "Test site",
                        "related": {},
                        "nav": [],
                    },
                    "type": "jinja2",
                },
            )

            rendered = page.render().content()
            self.assertEqual(b"", rendered)



from __future__ import annotations

import argparse
import logging
import sys
from typing import TYPE_CHECKING, Any

from . import cli
from .command import SiteCommand, register

if TYPE_CHECKING:
    import staticsite.site

log = logging.getLogger("site")


class FeatureCommand(cli.Command):
    """
    Command parser that can be run after the site is loaded
    """

    def __init__(self, site: staticsite.site.Site, args: argparse.Namespace):
        # Intentionally do not call super().__init__, to avoid reinitializing logs
        self.site = site
        self.args = args


class List(FeatureCommand):
    "list available features"

    def run(self) -> None:
        for feature in self.site.features.ordered():
            print(f"{feature.name} - {feature.get_short_description()}")


@register
class Site(SiteCommand):
    "run feature-specifc commands"

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "--cmd", nargs=argparse.REMAINDER, help="site-specific command (try 'help')"
        )
        return parser

    def run(self) -> None:
        site = self.load_site()

        # Build the site-specific command line parser
        parser = argparse.ArgumentParser(description="Site-specific commands.")
        subparsers = parser.add_subparsers(help="sub-command help", dest="command")

        List.add_subparser(subparsers)

        for feature in site.features.ordered():
            feature.add_site_commands(subparsers)

        args = parser.parse_args(self.args.cmd)

        if args.command is None:
            parser.print_help()
        else:
            handler = args.handler(site, args)
            res = handler.run()
            if res is not None:
                sys.exit(res)



from __future__ import annotations

import inspect
import logging
import os
from typing import TYPE_CHECKING, Any

from . import asset

if TYPE_CHECKING:
    from .feature import Feature
    from .fields import Field
    from .page import Page
    from .site import Site

log = logging.getLogger("autodoc")


def summary(obj: Any) -> str:
    if obj.__doc__ is None:
        return f"Missing documentation for {obj!r}"
    return inspect.cleandoc(obj.__doc__).strip().split("\n\n", 1)[0]


def body(obj: Any) -> str:
    if obj.__doc__ is None:
        return f"Missing documentation for {obj!r}"
    res = inspect.cleandoc(obj.__doc__).strip().split("\n\n", 1)
    if len(res) == 2:
        return res[1]
    else:
        return res[0]


class Autodoc:
    """
    Autogenerate reference documentation
    """

    def __init__(self, site: Site, root: str):
        self.site = site
        self.root = root

    def generate(self) -> None:
        # Iterate features to generate per-feature documentation
        page_types: set[type[Page]] = {asset.Asset}
        for feature in self.site.features.ordered():
            self.write_feature(feature)
            page_types.update(feature.get_used_page_types())

        # Iterate page types
        by_name: dict[str, type[Page]] = {}
        for base_page_type in page_types:
            page_type = self.site.features.get_page_class(base_page_type)
            name = page_type.TYPE
            if (old := by_name.get(name)) is not None:
                raise RuntimeError(
                    f"Page type {name} defined as both {old} and {page_type}"
                )
            else:
                by_name[name] = page_type
        fields: dict[str, Field[Any, Any]] = {}
        for name, page_type in by_name.items():
            self.write_page_type(name, page_type)
            fields.update(page_type._fields)

        # Iterate fields
        for name, field in fields.items():
            self.write_field(name, field)

        # Generate index
        with open(os.path.join(self.root, "README.md"), "wt") as out:
            print("# Reference documentation", file=out)
            print(file=out)
            print("## Core structure", file=out)
            print(file=out)
            print("* [The Site object](site.md)", file=out)
            # print("* [The Page object](page.md)", file=out)
            print("* [Site settings](settings.md)", file=out)
            print("* [Source contents](contents.md)", file=out)
            print("* [Themes](theme.md)", file=out)
            # print("* [Jinja2 templates reference](templates.md)", file=out)
            print("* [Selecting site pages](page-filter.md)", file=out)
            print("* [Archetypes reference](archetypes.md)", file=out)
            print("* [Site-specific features](feature.md)", file=out)
            print("* [Front Matter](front-matter.md)", file=out)
            print(file=out)

            print("## Features", file=out)
            print(file=out)
            for feature in sorted(self.site.features.ordered(), key=lambda f: f.name):
                print(
                    f"* [{feature.name}](features/{feature.name}.md): {summary(feature)}",
                    file=out,
                )
            print(file=out)

            print("## Page types", file=out)
            print(file=out)
            for name, page_type in sorted(by_name.items()):
                print(f"* [{name}](pages/{name}.md): {summary(page_type)}", file=out)
            print(file=out)

            print("## Page fields", file=out)
            print(file=out)
            for name, field in sorted(fields.items()):
                print(f"* [{name}](fields/{name}.md): {summary(field)}", file=out)
            print(file=out)
            print("[Back to README](../../README.md)", file=out)

        # - update self-generated metadata fields documentation
        #    - generate documentation for the various kinds of pages
        #    - lift __doc__ from pages, or another attribute gathered from all mixins
        #      by the metaclass, or by the documenter by following the MRO, for more
        #      documentation content

    def write_feature(self, feature: Feature) -> None:
        page_types = feature.get_used_page_types()
        path = os.path.join(self.root, "features")
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, f"{feature.name}.md"), "wt") as out:
            print(f"# {feature.name}: {summary(feature)}", file=out)
            print(file=out)

            if page_types:
                print("## Page types", file=out)
                print(file=out)
                for page_type in page_types:
                    print(
                        f"* [{page_type.TYPE}](../pages/{page_type.TYPE}.md): {summary(page_type)}",
                        file=out,
                    )
                print(file=out)

            print("## Documentation", file=out)
            print(file=out)
            print(body(feature), file=out)
            print(file=out)
            print("[Back to reference index](../README.md)", file=out)

    def write_page_type(self, name: str, page_type: type[Page]) -> None:
        if page_type.__doc__ is None:
            log.error("%s: page type is undocumented in %s", name, page_type)
            return
        path = os.path.join(self.root, "pages")
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, f"{name}.md"), "wt") as out:
            print(f"# {name}: {summary(page_type)}", file=out)
            print(file=out)
            print("## Fields", file=out)
            print(file=out)
            for name, field in sorted(page_type._fields.items()):
                print(f"* [{name}](../fields/{name}.md): {summary(field)}", file=out)
            print(file=out)
            print("## Documentation", file=out)
            print(file=out)
            print(body(page_type), file=out)
            print(file=out)
            print("[Back to reference index](../README.md)", file=out)

    def write_field(self, name: str, field: Field[Any, Any]) -> None:
        path = os.path.join(self.root, "fields")
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, f"{name}.md"), "wt") as out:
            print(f"# {name}: {summary(field)}", file=out)
            print(file=out)
            print(body(field), file=out)
            print(file=out)
            print("[Back to reference index](../README.md)", file=out)



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, Optional

from staticsite import fields
from staticsite.feature import Feature
from staticsite.page import ChangeExtent, Page, StandaloneAutoPage, TemplatePage
from staticsite.source_node import SourcePageNode

if TYPE_CHECKING:
    from .. import file
    from ..node import Node
    from ..site import Site

log = logging.getLogger("dirindex")


class ParentField(fields.Field["Dir", Optional[Page]]):
    """
    Field that works as a proxy for page.node.parent.page
    """

    def __get__(self, page: Page, type: type[Dir] | None = None) -> Page | None:
        if (parent := page.node.parent) is None:
            return None
        return parent.page


class Dir(TemplatePage, StandaloneAutoPage):
    """
    Page with a directory index.

    For each directory that does not have an index page, staticsite automatically
    generates one: this makes every partial url to the site a valid URL.

    Directory pages have these extra properties:

    * `page.meta.template` defaults to `dir.html`
    * `page.meta.title` is the directory name, or the site name in case the site
      root is an autogenerated directory page
    * `page.meta.pages` lists the pages in this directory
    * `page.meta.parent` parent page in the directory hierarchy
    """

    TYPE = "dir"
    TEMPLATE = "dir.html"

    parent = ParentField(doc="Page one level above in the site hierarchy")

    def __init__(self, site: Site, *, name: str | None = None, **kw: Any):
        super().__init__(site, **kw)
        # Directory name
        self.name: str | None = name
        # Subdirectory of this directory
        self.subdirs: list[Page] = []

        # Pointer to the directory in the file system
        self.src: file.File

        self.syndicated = False
        self.indexed = False

        if self.node.parent:
            self.title = self.name

        pages: list[Page] = []
        for name, sub in self.node.sub.items():
            if sub.page:
                if sub.page.directory_index:
                    self.subdirs.append(sub.page)
                else:
                    pages.append(sub.page)
        for name, page in self.node.build_pages.items():
            if page != self and page.leaf and page.indexed:
                pages.append(page)
        pages.sort(key=lambda p: p.date)
        self.pages = pages

        # self.indexed = bool(self.pages) or any(p.indexed for p in self.subdirs)

        # Since directory indices is called from the bottom up, subdirs have
        # their date up to date
        self.subdirs.sort(key=lambda p: p.date)

    def crossreference(self) -> None:
        date_pages = []
        if self.subdirs:
            date_pages.append(self.subdirs[-1].date)
        if self.pages:
            date_pages.append(self.pages[-1].date)

        if date_pages:
            self.date = max(date_pages)
        else:
            self.date = self.site.localized_timestamp(self.src.stat.st_mtime)

    def _compute_change_extent(self) -> ChangeExtent:
        res = super()._compute_change_extent()

        # Check if pages were deleted in this dir
        for relpath in self.site.deleted_source_pages():
            if os.path.dirname(relpath) == self.src.relpath:
                return ChangeExtent.ALL

        # Dir has changed if any page referenced changed in metadata
        for subdir in self.subdirs:
            if subdir.change_extent == ChangeExtent.ALL:
                res = ChangeExtent.ALL
        if self.pages:
            for page in self.pages:
                if page.change_extent == ChangeExtent.ALL:
                    res = ChangeExtent.ALL
        return res


class DirindexFeature(Feature):
    """
    Build redirection pages for page aliases.

    A page can define 'aliases=[...]' to generate pages in those locations that
    redirect to the page.
    """

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.dir_pages: list[Dir] = []

    def get_used_page_types(self) -> list[type[Page]]:
        return [Dir]

    def scan(self, node: Node) -> None:
        for subnode in node.sub.values():
            self.scan(subnode)

        if isinstance(node, SourcePageNode) and not node.page:
            page = node.create_auto_page_as_index(page_cls=Dir, name=node.name)
            if page is not None:
                page.src = node.src
                self.dir_pages.append(page)

    def generate(self) -> None:
        # Scan the node hierarchy creating indices for nodes without an index
        # page
        self.scan(self.site.root)

    def crossreference(self) -> None:
        for page in self.dir_pages:
            page.crossreference()


FEATURES = {
    "dirindex": DirindexFeature,
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .page import SourcePage
from .render import RenderedElement, RenderedFile

if TYPE_CHECKING:
    from .site import Site


class Asset(SourcePage):
    """
    A static asset, copied as is
    """

    TYPE = "asset"

    def __init__(self, site: Site, *, name: str, **kw: Any):
        super().__init__(site, **kw)
        self.date = site.localized_timestamp(self.src.stat.st_mtime)
        self.title = name
        self.asset = True
        self.draft = False
        self.indexed = False
        self.name = name
        self.ready_to_render = True

    def render(self, **kw: Any) -> RenderedElement:
        return RenderedFile(self.src)



from __future__ import annotations

from unittest import TestCase

from . import utils as test_utils


class TestAliases(test_utils.MockSiteTestMixin, TestCase):
    def test_site(self):
        files = {
            "page.md": {
                "aliases": ["alias", "test/alias"],
                "title": "Page Title",
            },
        }
        with self.site(files) as mocksite:
            page, alias1, alias2 = mocksite.page("page", "alias", "test/alias")

            self.assertEqual(page.node.path, "page")
            self.assertEqual(page.title, "Page Title")
            self.assertEqual(page.node.path, "page")
            self.assertEqual(page.dst, "index.html")
            self.assertIsNotNone(page.aliases)

            self.assertEqual(alias1.node.path, "alias")
            self.assertEqual(alias1.title, "Page Title")
            self.assertEqual(alias1.node.path, "alias")
            self.assertEqual(alias1.dst, "index.html")
            self.assertEqual(alias1.template, "redirect.html")
            self.assertEqual(alias1.page, page)
            self.assertIsNone(alias1.aliases)

            self.assertEqual(alias2.node.path, "test/alias")
            self.assertEqual(alias2.title, "Page Title")
            self.assertEqual(alias2.node.path, "test/alias")
            self.assertEqual(alias2.dst, "index.html")
            self.assertEqual(alias2.template, "redirect.html")
            self.assertIsNone(alias2.aliases)

    def test_conflict(self):
        self.maxDiff = None
        files = {
            "page.md": {"aliases": ["alias", "/page", "/dir1/dir2/alias"]},
        }
        with self.site(files, auto_load_site=False) as mocksite:
            with self.assertLogs(level="WARNING") as out:
                mocksite.load_site()
            self.assertEqual(
                out.output,
                [
                    "WARNING:aliases:markdown:page.md defines alias '/page' pointing to itself"
                ],
            )
            mocksite.assertPagePaths(("", "page", "alias", "dir1/dir2/alias"))
            page, alias, alias1 = mocksite.page("page", "alias", "dir1/dir2/alias")

            self.assertEqual(page.TYPE, "markdown")
            self.assertEqual(page.node.path, "page")
            self.assertEqual(page.dst, "index.html")
            self.assertEqual(page.aliases, ["alias", "/page", "/dir1/dir2/alias"])

            self.assertEqual(alias.node.path, "alias")
            self.assertEqual(alias.dst, "index.html")
            self.assertEqual(alias.template, "redirect.html")
            self.assertEqual(alias.page, page)
            self.assertIsNone(alias.aliases)
            mocksite.assertBuilt(
                None, "alias", "alias/index.html", sample='href="/page'
            )

            self.assertEqual(alias1.node.path, "dir1/dir2/alias")
            self.assertEqual(alias1.dst, "index.html")
            self.assertEqual(alias1.template, "redirect.html")
            self.assertEqual(alias1.page, page)
            self.assertIsNone(alias.aliases)
            mocksite.assertBuilt(
                None,
                "dir1/dir2/alias",
                "dir1/dir2/alias/index.html",
                sample='href="/page',
            )



import os
from unittest import TestCase

from . import utils as test_utils


class TestMetadata(test_utils.MockSiteTestMixin, TestCase):
    """
    Test metadata collected on site load
    """

    def test_dir(self):
        files = {
            ".staticsite": {
                "files": {
                    "*.html": {
                        "title": "Default title",
                    },
                    "test.html": {
                        "title": "Test title",
                    },
                },
            },
            "index.html": "",
            "test.html": "",
            "test1.html": """
{% block front_matter %}
---
title: Test1 title
{% endblock %}""",
        }

        with self.site(files) as mocksite:
            mocksite.assertPagePaths(("", "test.html", "test1.html"))
            index, test, test1 = mocksite.page("", "test.html", "test1.html")

            self.assertEqual(index.TYPE, "jinja2")
            self.assertEqual(test.TYPE, "jinja2")
            self.assertEqual(test1.TYPE, "jinja2")

            self.assertEqual(index.title, "Default title")
            self.assertEqual(test.title, "Test title")
            self.assertEqual(test1.title, "Test1 title")

    def test_tree_meta(self):
        files = {
            ".staticsite": {
                "site_name": "Root site",
            },
            "page.md": {},
            "page1.md": {"site_name": "Page 1 site"},
            "dir1/page.md": {},
            "dir1/dir2/.staticsite": {
                "site_name": "dir2 site",
            },
            "dir1/dir2/page.md": {},
        }

        with self.site(files) as mocksite:
            self.assertEqual(mocksite.page("").site_name, "Root site")
            self.assertEqual(mocksite.page("page").site_name, "Root site")
            self.assertEqual(mocksite.page("page1").site_name, "Page 1 site")
            self.assertEqual(mocksite.page("dir1").site_name, "Root site")
            self.assertEqual(mocksite.page("dir1/page").site_name, "Root site")
            self.assertEqual(mocksite.page("dir1/dir2").site_name, "dir2 site")
            self.assertEqual(mocksite.page("dir1/dir2/page").site_name, "dir2 site")

    def test_asset(self):
        self.maxDiff = None

        files = {
            ".staticsite": {
                "files": {
                    "test.md": {
                        "asset": True,
                    },
                },
            },
            "test.md": {},
            "test1.md": {},
        }

        with self.site(files) as mocksite:
            mocksite.assertPagePaths(("", "test.md", "test1"))
            index, test, test1 = mocksite.page("", "test.md", "test1")
            self.assertEqual(index.TYPE, "dir")
            self.assertEqual(test.TYPE, "asset")
            self.assertEqual(test1.TYPE, "markdown")

    def test_dir_asset(self):
        files = {
            ".staticsite": {
                "dirs": {
                    "examples": {
                        "asset": True,
                    },
                },
            },
            "test.md": {},
            "examples/test1.md": {},
            "examples/subdir/test2.md": {},
        }

        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                ("", "test", "examples/test1.md", "examples/subdir/test2.md")
            )
            index, test, test1, test2 = mocksite.page(
                "", "test", "examples/test1.md", "examples/subdir/test2.md"
            )
            self.assertEqual(index.TYPE, "dir")
            self.assertEqual(test.TYPE, "markdown")
            self.assertEqual(test1.TYPE, "asset")
            self.assertEqual(test2.TYPE, "asset")


class TestSiteName(test_utils.MockSiteTestMixin, TestCase):
    def test_from_content_dir_name(self):
        sitedef = test_utils.MockSite(
            {
                "index.md": {},
                "page.md": {},
                "dir/page.md": {},
            }
        )
        sitedef.settings.SITE_NAME = None

        with self.site(sitedef) as mocksite:
            expected = os.path.basename(mocksite.root)
            self.assertEqual(mocksite.page("").site_name, expected)
            self.assertEqual(mocksite.page("page").site_name, expected)
            self.assertEqual(mocksite.page("dir/page").site_name, expected)

    def test_from_settings(self):
        sitedef = test_utils.MockSite(
            {
                "index.md": {},
                "page.md": {},
                "dir/page.md": {},
            }
        )
        sitedef.settings.SITE_NAME = "Site Name"

        with self.site(sitedef) as mocksite:
            self.assertEqual(mocksite.page("").site_name, "Site Name")
            self.assertEqual(mocksite.page("page").site_name, "Site Name")
            self.assertEqual(mocksite.page("dir/page").site_name, "Site Name")

    def test_from_dir_meta(self):
        files = {
            ".staticsite": {"site_name": "Site Name dirmeta"},
            "index.md": {},
            "page.md": {},
            "dir/page.md": {},
        }

        with self.site(files) as mocksite:
            self.assertEqual(mocksite.page("").site_name, "Site Name dirmeta")
            self.assertEqual(mocksite.page("page").site_name, "Site Name dirmeta")
            self.assertEqual(mocksite.page("dir/page").site_name, "Site Name dirmeta")

    def test_from_root_title(self):
        sitedef = test_utils.MockSite(
            {
                "index.md": {"title": "Site Name title"},
                "page.md": {},
                "dir/page.md": {},
            }
        )
        sitedef.settings.SITE_NAME = None

        with self.site(sitedef) as mocksite:
            self.assertEqual(mocksite.page("").site_name, "Site Name title")
            self.assertEqual(mocksite.page("page").site_name, "Site Name title")
            self.assertEqual(mocksite.page("dir/page").site_name, "Site Name title")


class TestFields(test_utils.MockSiteTestMixin, TestCase):
    def test_date(self):
        self.maxDiff = None

        sitedef = test_utils.MockSite(
            {
                "index.html": """
{% block front_matter %}
---
title: Test1 title
date: 2000-01-01 00:00:00
{% endblock %}""",
                "page.md": """---
date: 2005-01-01 00:00:00
---
""",
                "page1.yaml": """---
data_type: test
date: 2010-01-01 00:00:00
""",
            }
        )
        sitedef.settings.SITE_NAME = None
        sitedef.settings.TIMEZONE = "UTC"

        with self.site(sitedef) as mocksite:
            j2page, mdpage, datapage = mocksite.page("", "page", "page1")

            self.assertEqual(
                j2page.to_dict(),
                {
                    "src": {
                        "relpath": "index.html",
                        "abspath": os.path.join(mocksite.root, "index.html"),
                    },
                    "site_path": "",
                    "build_path": "index.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2000 Test User",
                        "date": "2000-01-01 00:00:00+00:00",
                        "draft": False,
                        "indexed": True,
                        "syndicated": True,
                        "syndication_date": "2000-01-01 00:00:00+00:00",
                        "nav": [],
                        "site_name": "Test1 title",
                        "site_url": "https://www.example.org",
                        "template": "compiled:index.html",
                        "template_copyright": "compiled:None",
                        "title": "Test1 title",
                        "related": {},
                    },
                    "type": "jinja2",
                },
            )

            self.assertEqual(
                mdpage.to_dict(),
                {
                    "src": {
                        "relpath": "page.md",
                        "abspath": os.path.join(mocksite.root, "page.md"),
                    },
                    "site_path": "page",
                    "build_path": "page/index.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2005 Test User",
                        "date": "2005-01-01 00:00:00+00:00",
                        "draft": False,
                        "indexed": True,
                        "syndicated": True,
                        "syndication_date": "2005-01-01 00:00:00+00:00",
                        "site_name": "Test1 title",
                        "site_url": "https://www.example.org",
                        "template": "page.html",
                        "template_copyright": "compiled:None",
                        "title": "Test1 title",
                        "related": {},
                        "nav": [],
                    },
                    "type": "markdown",
                },
            )

            self.assertEqual(
                datapage.to_dict(),
                {
                    "src": {
                        "relpath": "page1.yaml",
                        "abspath": os.path.join(mocksite.root, "page1.yaml"),
                    },
                    "site_path": "page1",
                    "build_path": "page1/index.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2010 Test User",
                        "date": "2010-01-01 00:00:00+00:00",
                        "draft": False,
                        "indexed": True,
                        "syndicated": True,
                        "syndication_date": "2010-01-01 00:00:00+00:00",
                        "site_name": "Test1 title",
                        "site_url": "https://www.example.org",
                        "template": "compiled:data.html",
                        "template_copyright": "compiled:None",
                        "title": "Test1 title",
                        "data_type": "test",
                        "related": {},
                        "nav": [],
                    },
                    "type": "data",
                },
            )



from __future__ import annotations

import io
import logging
import os
import re
from collections import defaultdict
from collections.abc import Sequence
from typing import IO, TYPE_CHECKING, Any

import jinja2

from staticsite import fields
from staticsite.archetypes import Archetype
from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.features.jinja2 import RenderPartialTemplateMixin
from staticsite.page import Page, SourcePage, TemplatePage
from staticsite.page_filter import PageFilter
from staticsite.utils import yaml_codec

if TYPE_CHECKING:
    from staticsite import file, fstree
    from staticsite.archetypes import Archetypes
    from staticsite.site import Site
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("data")


re_ext = re.compile(r"\.(json|toml|yaml)$")


class DataTypeField(TrackedField["DataPageMixin", str], fields.Str["DataPageMixin"]):
    """
    Type of data for this file.

    Identifies the data type. Internally, the data feature groups data pages by
    type, so further features can efficiently access thematic datasets.

    This is used to group data of the same type together, and to choose a
    `data-[data_type].html` rendering template.

    The `page.meta.template` metadata for data pages, when not specified, defaults
    to `dir-[type].html`, or if that is missing, to `data.html`.
    """

    tracked_by = "data"


class DataPageMixin(Page):
    data_type = DataTypeField()


class DataPages(PageTrackingMixin["DataPageMixin"], Feature):
    """
    Handle datasets in content directories.

    This allows storing pure-data datasets in JSON, Yaml, or Toml format in
    the contents directory, access it from pages, and render it using Jinja2
    templates.

    Each dataset needs, at its toplevel, to be a dict with a ``type`` element,
    and the dataset will be rendered using the ``data-{{type}}.html`` template.

    Other front-matter attributes like ``date``, ``title``, ``aliases``, and
    taxonomy names are handled as with other pages. The rest of the dictionary
    is ignored and can contain any data one wants.
    """

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.by_type: dict[str, list[Page]] = defaultdict(list)
        self.j2_globals["data_pages"] = self.jinja2_data_pages
        self.page_class_by_type: dict[str, type[Page]] = {}

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return (DataPageMixin,)

    def register_page_class(self, type: str, cls: type[Page]) -> None:
        self.page_class_by_type[type] = cls

    def get_used_page_types(self) -> list[type[Page]]:
        return [DataPage]

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            mo = re_ext.search(fname)
            if not mo:
                continue
            taken.append(fname)

            fmt = mo.group(1)

            with directory.open(fname, "rt") as fd:
                try:
                    fm_meta = parse_data(fd, fmt)
                except Exception:
                    log.exception("%s: failed to parse %s content", src.relpath, fmt)
                    continue

            try:
                data_type = fm_meta.get("data_type", None)
            except AttributeError:
                log.error("%s: data did not parse into a dict", src.relpath)
                continue

            if data_type is None:
                log.error("%s: data_type not found: ignoring page", src.relpath)
                continue

            page_name = fname[: -len(mo.group(0))]
            kwargs.update(fm_meta)
            kwargs["page_cls"] = self.page_class_by_type.get(data_type, DataPage)
            kwargs["src"] = src

            if page_name == "index":
                page = node.create_source_page_as_index(**kwargs)
            else:
                page = node.create_source_page_as_path(name=page_name, **kwargs)
            if page is not None:
                pages.append(page)

        for fname in taken:
            del files[fname]

        return pages

    def try_load_archetype(
        self, archetypes: Archetypes, relpath: str, name: str
    ) -> Archetype | None:
        mo = re_ext.search(relpath)
        if not mo:
            return None
        fmt = mo.group(1)
        if os.path.basename(relpath) != name + "." + fmt:
            return None
        return DataArchetype(archetypes, relpath, self, fmt)

    def organize(self) -> None:
        # Dispatch pages by type
        for page in self.tracked_pages:
            data_type = page.data_type
            if data_type is None:
                continue
            self.by_type[data_type].append(page)

        # Sort the pages of each type by date
        for pages in self.by_type.values():
            pages.sort(key=lambda x: x.date)

    @jinja2.pass_context
    def jinja2_data_pages(
        self,
        context: jinja2.runtime.Context,
        type: str,
        path: str | None = None,
        limit: int | None = None,
        sort: str | None = None,
        **kw: str,
    ) -> list[Page]:
        page_filter = PageFilter(
            self.site,
            path=path,
            limit=limit,
            sort=sort,
            allow=self.by_type.get(type, []),
            root=None,
            **kw,
        )
        return page_filter.filter()


def parse_data(fd: IO[str], fmt: str) -> Any:
    if fmt == "json":
        import json

        return json.load(fd)
    elif fmt == "toml":
        import toml

        return toml.load(fd)
    elif fmt == "yaml":
        return yaml_codec.load(fd)
    else:
        raise NotImplementedError(f"data format {fmt} is not supported")


def write_data(fd: IO[str], data: dict[str, Any], fmt: str) -> None:
    if fmt == "json":
        import json

        json.dump(data, fd)
    elif fmt == "toml":
        import toml

        toml.dump(data, fd)
    elif fmt == "yaml":
        yaml_codec.dump(data, fd)
    else:
        raise NotImplementedError(f"data format {fmt} is not supported")


class DataPage(RenderPartialTemplateMixin, TemplatePage, SourcePage):
    """
    Data files

    Data files have a `.json`, `.yaml`, or `.toml` extension and can be rendered
    with custom Jinja2 templates.

    The content of the data file is parsed, and merged into the page metadata.

    The jinja2 template can be chosen using the `data_type` metadata field.

    The metadata of any page with the `data_type` metadata will be also tracked as
    data. This allows to create normal pages that also add to a named dataset.
    """

    TYPE = "data"

    def __init__(self, site: Site, **kw: Any):
        # Indexed by default
        kw.setdefault("indexed", True)
        if "template" not in kw:
            kw["template"] = site.theme.jinja2.select_template(
                [f"data-{kw['data_type']}.html", "data.html"]
            )
        super().__init__(site, **kw)


class DataArchetype(Archetype):
    def __init__(
        self, archetypes: Archetypes, relpath: str, data_pages: DataPages, fmt: str
    ):
        super().__init__(archetypes, relpath)
        self.data_pages = data_pages
        self.format = fmt

    def render(self, **kw: Any) -> tuple[dict[str, Any], str]:
        meta, rendered = super().render(**kw)

        # Reparse the rendered version
        with io.StringIO(rendered) as fd:
            data = parse_data(fd, self.format)

        # Make a copy of the full parsed metadata
        archetype_meta = dict(data)

        # Remove the path entry
        data.pop("path", None)

        # Reserialize the data
        with io.StringIO() as fd:
            write_data(fd, data, self.format)
            post_body = fd.getvalue()

        return archetype_meta, post_body


FEATURES = {
    "data": DataPages,
}



from __future__ import annotations

import argparse
import logging
import sys
from typing import Any, cast

try:
    import coloredlogs

    HAS_COLOREDLOGS = True
except ModuleNotFoundError:
    HAS_COLOREDLOGS = False


def _get_first_docstring_line(obj: Any) -> str | None:
    if obj.__doc__ is None:
        raise RuntimeError(f"{obj!r} lacks a docstring")
    try:
        return cast(str, obj.__doc__).split("\n")[1].strip()
    except (AttributeError, IndexError):
        return None


class Fail(BaseException):
    """
    Failure that causes the program to exit with an error message.

    No stack trace is printed.
    """


class Success(BaseException):
    """
    Exception raised when a command has been successfully handled, and no
    further processing should happen
    """


class Command:
    """
    Base class for actions run from command line
    """

    NAME: str

    def __init__(self, args: argparse.Namespace):
        self.NAME = getattr(self.__class__, "NAME", self.__class__.__name__.lower())
        self.args = args
        self.setup_logging()

    def setup_logging(self) -> None:
        FORMAT = "%(asctime)-15s %(levelname)s %(name)s %(message)s"
        if self.args.debug:
            level = logging.DEBUG
        elif self.args.verbose:
            level = logging.INFO
        else:
            level = logging.WARN

        if HAS_COLOREDLOGS:
            coloredlogs.install(level=level, fmt=FORMAT)
        else:
            logging.basicConfig(level=level, stream=sys.stderr, format=FORMAT)

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        cls.NAME = getattr(cls, "NAME", cls.__name__.lower())
        parser: argparse.ArgumentParser = subparsers.add_parser(
            cls.NAME,
            help=_get_first_docstring_line(cls),
        )
        parser.set_defaults(command=cls)
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="verbose output",
        ),
        parser.add_argument(
            "--debug",
            action="store_true",
            help="debugging output",
        ),
        return parser



#!/usr/bin/python3

import timeit
from staticsite import Site, Settings


def setup():
    settings = Settings()
    settings.PROJECT_ROOT = "example"
    settings.load("example/settings.py")
    site = Site(settings)
    site.features.load_default_features()
    site.load_theme()
    # Patch add_page to be a noop
    site.add_page = lambda page: None
    return site


val = timeit.timeit("site.load_content()", setup="site = setup()\nsite.load_content()", number=300, globals=globals())
print(f"Current: {val}")



from __future__ import annotations

import datetime
import logging
from collections import Counter
from typing import TYPE_CHECKING

from staticsite.utils import timings

from .command import SiteCommand, register

if TYPE_CHECKING:
    from ..site import Site

log = logging.getLogger("check")


@register
class Check(SiteCommand):
    "check the site, going through all the motions of rendering it without writing anything"

    def run(self) -> None:
        site = self.load_site()
        with timings("Checked site in %fs"):
            self.check(site)

    def check(self, site: Site) -> None:
        counts: dict[str, int] = Counter()
        for page in site.iter_pages():
            counts[page.TYPE] += 1
            page.check()

            date = page.meta.get("date", None)
            if date is not None:
                if not isinstance(date, datetime.datetime):
                    log.info(
                        "%r: meta date %r is not an instance of datetime", page, date
                    )

        for type, count in sorted(counts.items()):
            print(f"{count} {type} pages")



from __future__ import annotations

import logging
import os
from typing import Any

from .command import Fail, SiteCommand, register

log = logging.getLogger("shell")


@register
class Shell(SiteCommand):
    "start a shell with the global `site` set to the current site"

    # Taken from django's core/management/commands/shell.py

    shells = ["ipython", "bpython", "python"]

    def ipython(self, **kw: Any) -> None:
        from IPython import start_ipython

        start_ipython(argv=[], user_ns=kw)

    def bpython(self, **kw: Any) -> None:
        import bpython

        bpython.embed(locals_=kw)

    def python(self, **kw: Any) -> None:
        import code

        # Set up a dictionary to serve as the environment for the shell, so
        # that tab completion works on objects that are imported at runtime.
        imported_objects = kw
        try:  # Try activating rlcompleter, because it's handy.
            import readline
        except ImportError:
            pass
        else:
            # We don't have to wrap the following import in a 'try', because
            # we already know 'readline' was imported successfully.
            import rlcompleter

            readline.set_completer(rlcompleter.Completer(imported_objects).complete)
            # Enable tab completion on systems using libedit (e.g. macOS).
            # These lines are copied from Lib/site.py on Python 3.4.
            readline_doc = getattr(readline, "__doc__", "")
            if readline_doc is not None and "libedit" in readline_doc:
                readline.parse_and_bind("bind ^I rl_complete")
            else:
                readline.parse_and_bind("tab:complete")

        # We want to honor both $PYTHONSTARTUP and .pythonrc.py, so follow system
        # conventions and get $PYTHONSTARTUP first then .pythonrc.py.
        for pythonrc in {
            os.environ.get("PYTHONSTARTUP"),
            os.path.expanduser("~/.pythonrc.py"),
        }:
            if not pythonrc:
                continue
            if not os.path.isfile(pythonrc):
                continue
            try:
                with open(pythonrc) as handle:
                    exec(compile(handle.read(), pythonrc, "exec"), imported_objects)
            except NameError:
                pass
        code.interact(local=imported_objects)

    def run(self) -> None:
        site = self.load_site()

        for shell in self.shells:
            try:
                getattr(self, shell)(site=site)
            except ModuleNotFoundError:
                pass

        raise Fail(f"Could not import {shell} interface.")



#!/usr/bin/python3

import timeit
from staticsite.utils import yaml_codec  # noqa

test_yaml = """
---
# In staticsite, a taxonomy is a group of attributes like categories or tags.
#
# Like in Hugo, you can have as many taxonomies as you want. See
# https://gohugo.io/taxonomies/overview/ for a general introduction to
# taxonomies.
#
# This file describes the taxonomy for "tags". The name of the taxonomy is
# taken from the file name.
#
# The format of the file is the same that is used for the front matter of
# posts, again same as in Hugo: https://gohugo.io/content/front-matter/

# Any toplevel metadata is used for the tag index
title: "All series"
description: "Index of all series in the site."
# template: tags.html

# Category metadata is used for each tag page
category:
  # template: tag.html
  template_title: "Latest posts of series <strong>{{page.name}}</strong>"
  template_description: "Most recent posts of series <strong>{{page.name}}</strong>"
  syndication:
    template_title: "{{page.meta.site_name}}: posts of series {{page.meta.index.name}}"
    template_description: "{{page.meta.site_name}}: most recent posts of series {{page.meta.index.name}}"

# Archive metadata is used for each tag archive page
archive:
  # template: tag-archive.html
  template_title: "Archive of posts of series <strong>{{page.name}}</strong>"
  template_description: "Archive of all posts of series <strong>{{page.name}}</strong>"
"""

val = timeit.timeit("yaml_codec.loads_ruamel(test_yaml)", number=1000, globals=globals())
print(f"Ruamel: {val}")

val = timeit.timeit("yaml_codec.loads_pyyaml(test_yaml)", number=1000, globals=globals())
print(f"PyYAML: {val}")



from __future__ import annotations

import os
from unittest import TestCase

from . import utils as test_utils


class TestDemo(test_utils.SiteTestMixin, TestCase):
    site_name = "demo"
    site_settings = {"SITE_AUTHOR": "Test User"}

    @test_utils.assert_no_logs()
    def test_meta(self):
        page = self.page("")
        self.assertEqual(page.site_path, "")
        self.assertEqual(page.meta["site_name"], "Example web site")
        self.assertEqual(page.meta["site_url"], "https://www.example.org")
        self.assertEqual(page.meta["author"], "Example author")

        page = self.page("blog")
        self.assertEqual(page.site_path, "blog")
        self.assertEqual(page.meta["site_name"], "Example web site")
        self.assertEqual(page.meta["site_url"], "https://www.example.org")
        self.assertEqual(page.meta["author"], "Example author")

    @test_utils.assert_no_logs()
    def test_dots(self):
        output = os.path.join(self.build_root, "index.html")
        self.assertTrue(os.path.exists(output))
        output = os.path.join(self.build_root, ".secret")
        self.assertFalse(os.path.exists(output))
        output = os.path.join(self.build_root, ".secrets")
        self.assertFalse(os.path.exists(output))

    @test_utils.assert_no_logs()
    def test_render_paths(self):
        self.assertBuilt("index.html", "", "index.html")
        self.assertBuilt("pages/README.md", "pages", "pages/index.html")
        self.assertBuilt(
            "pages/doc/sub.md", "pages/doc/sub", "pages/doc/sub/index.html"
        )

    @test_utils.assert_no_logs()
    def test_different_links(self):
        # Check rendered pages
        page = self.page("pages")
        self.assertEqual(
            page.body_start, ["Link: [Docs](doc/sub.md)", "", "[Back home](..)"]
        )

        output = os.path.join(self.build_root, "pages/index.html")
        with open(output, encoding="utf8") as fd:
            content = fd.read()
        self.assertIn('<a href="/pages/doc/sub">Docs</a>', content)
        self.assertIn('<a href="/">Back home</a>', content)

        output = os.path.join(self.build_root, "pages/doc/sub/index.html")
        with open(output, encoding="utf8") as fd:
            content = fd.read()
        self.assertIn('<a href="/pages">Back to README</a>', content)

    def test_titles(self):
        self.maxDiff = None
        page = self.page("tags/example")
        self.assertIsNotNone(page.meta["template_title"])
        self.assertEqual(
            page.meta["title"], "Latest posts for tag <strong>example</strong>"
        )

        self.assertEqual(
            self.page("blog/index.rss").to_dict(),
            {
                "site_path": "blog/index.rss",
                "build_path": "blog/index.rss",
                "meta": {
                    "date": "2016-04-16 10:23:00+02:00",
                    "syndication_date": "2016-04-16 10:23:00+02:00",
                    "index": "J2Page(blog)",
                    "author": "Example author",
                    "copyright": "© 2016 Example author",
                    "created_from": "J2Page(blog)",
                    "indexed": False,
                    "syndicated": False,
                    "nav": [],
                    "series": [],
                    "tags": [],
                    "pages": [
                        "RstPage(blog/2016/rst_example)",
                        "MarkdownPage(blog/2016/example-series3)",
                        "MarkdownPage(blog/2016/example-series2)",
                        "MarkdownPage(blog/2016/example-series1)",
                        "MarkdownPage(blog/2016/example)",
                    ],
                    "site_name": "Example web site",
                    "site_url": "https://www.example.org",
                    "template": "syndication.rss",
                    "template_copyright": "compiled:None",
                    "title": "Example blog feed",
                    "related": {},
                },
                "type": "rss",
            },
        )



from __future__ import annotations

import io
import logging
import os
import re
from typing import IO, TYPE_CHECKING, Any, BinaryIO, cast

import jinja2
import markdown
import markdown.treeprocessors
import markupsafe
from markdown.util import AMP_SUBSTITUTE

from staticsite.archetypes import Archetype
from staticsite.feature import Feature
from staticsite.markup import MarkupFeature, MarkupPage
from staticsite.page import FrontMatterPage, ImagePage, Page, TemplatePage
from staticsite.utils import front_matter

if TYPE_CHECKING:
    import xml.etree.ElementTree as ET

    from staticsite import file, fstree
    from staticsite.archetypes import Archetypes
    from staticsite.markup import LinkResolver
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("markdown")


class FixURLs(markdown.treeprocessors.Treeprocessor):
    """
    Markdown Treeprocessor that fixes internal links in HTML tags
    """

    def __init__(self, *args: Any, link_resolver: LinkResolver, **kw: Any) -> None:
        super().__init__(*args, **kw)
        self.link_resolver = link_resolver

    def should_resolve(self, url: str) -> bool:
        if url.startswith(AMP_SUBSTITUTE):
            # Possibly an overencoded mailto: link.
            # see https://bugs.debian.org/816218
            #
            # Markdown then further escapes & with utils.AMP_SUBSTITUTE, so
            # we look for it here.
            return False
        return True

    def run(self, root: ET.ElementTree) -> None:
        # Replace <a href=…>
        for a in root.iter("a"):
            if (orig_url := a.attrib.get("href", None)) is None:
                continue

            if not self.should_resolve(orig_url):
                continue

            new_url = self.link_resolver.resolve_url(orig_url)
            if new_url is not None:
                a.attrib["href"] = new_url

        # Replace <img src=…>
        for img in root.iter("img"):
            if (orig_url := img.attrib.get("src", None)) is None:
                continue

            if not self.should_resolve(orig_url):
                continue

            if (resolved := self.link_resolver.resolve_page(orig_url)) is None:
                continue

            if isinstance(resolved.page, ImagePage):
                attrs = resolved.page.get_img_attributes(
                    absolute=self.link_resolver.absolute
                )
            else:
                log.warning(
                    "%s: img src= resolves to %s which is not an image page",
                    self.link_resolver.page,
                    resolved.page,
                )
                continue

            img.attrib.update(attrs)


class StaticSiteExtension(markdown.extensions.Extension):
    def __init__(self, *, link_resolver: LinkResolver, **kwargs: Any):
        super().__init__(**kwargs)
        self.link_resolver = link_resolver

    def extendMarkdown(self, md: markdown.Markdown) -> None:
        md.treeprocessors.register(
            FixURLs(md, link_resolver=self.link_resolver), "staticsite", 0
        )
        md.registerExtension(self)

    def reset(self) -> None:
        pass

    def set_page(self, page: Page, absolute: bool) -> None:
        self.link_resolver.set_page(page, absolute)


class MarkdownPages(MarkupFeature, Feature):
    """
    Render ``.md`` markdown pages, with front matter.

    See doc/reference/markdown.md for details.
    """

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.markdown = markdown.Markdown(
            extensions=self.site.settings.MARKDOWN_EXTENSIONS
            + [
                StaticSiteExtension(link_resolver=self.link_resolver),
            ],
            extension_configs=self.site.settings.MARKDOWN_EXTENSION_CONFIGS,
            output_format="html",
        )

        self.j2_filters["markdown"] = self.jinja2_markdown

        self.render_cache = self.site.caches.get("markdown")

    def get_used_page_types(self) -> list[type[Page]]:
        return [MarkdownPage]

    @jinja2.pass_context
    def jinja2_markdown(self, context: jinja2.runtime.Context, mdtext: str) -> str:
        return markupsafe.Markup(self.render_snippet(context.parent["page"], mdtext))

    def render_snippet(self, page: Page, content: str) -> str:
        """
        Render markdown in the context of the given page.

        It renders the page content by default, unless `content` is set to a
        different markdown string.
        """
        self.link_resolver.set_page(page, absolute=False)
        self.markdown.reset()
        return self.markdown.convert(content)

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            if not fname.endswith(".md"):
                continue
            taken.append(fname)

            try:
                fm_meta, body = self.load_file_meta(directory, fname)
            except Exception as e:
                log.warning(
                    "%s: Failed to parse markdown page front matter (%s): skipped",
                    src,
                    e,
                )
                log.debug(
                    "%s: Failed to parse markdown page front matter: skipped",
                    src,
                    exc_info=e,
                )
                continue

            kwargs.update(fm_meta)
            kwargs["feature"] = self
            kwargs["src"] = src
            kwargs["page_cls"] = MarkdownPage
            kwargs["front_matter"] = fm_meta
            kwargs["body"] = body

            page: Page | None
            if fname in ("index.md", "README.md"):
                page = node.create_source_page_as_index(**kwargs)
            else:
                page = node.create_source_page_as_path(name=fname[:-3], **kwargs)

            if page is not None:
                pages.append(page)

        for fname in taken:
            del files[fname]

        return pages

    def load_dir_meta(self, directory: fstree.Tree) -> dict[str, Any] | None:
        # Load front matter from index.md
        # Do not try to load front matter from README.md, as one wouldn't
        # clutter a repo README with staticsite front matter
        if (src := directory.files.get("index.md")) is None:
            return None

        try:
            meta, body = self.load_file_meta(directory, "index.md")
        except Exception as e:
            log.debug("%s: failed to parse front matter", src.relpath, exc_info=e)
            log.warning("%s: failed to parse front matter", src.relpath)
        else:
            return meta

        return None

    def load_file_meta(
        self, directory: fstree.Tree, fname: str
    ) -> tuple[dict[str, Any], list[str]]:
        """
        Load metadata for a file.

        Returns the metadata and the markdown lines for the rest of the file
        """
        # Read the contents

        # Parse separating front matter and markdown content
        with directory.open(fname, "rb") as fd:
            return self.read_file_meta(fd)

    def read_file_meta(self, fd: IO[bytes]) -> tuple[dict[str, Any], list[str]]:
        """
        Load metadata for a file.

        Returns the metadata and the markdown lines for the rest of the file
        """
        fmt, meta, body = front_matter.read_markdown_partial(fd)

        body = list(body)

        # Remove leading empty lines
        while body and not body[0]:
            body.pop(0)

        # Read title from first # title if not specified in metadata
        if not meta.get("title", ""):
            if body and body[0].startswith("# "):
                meta["title"] = body.pop(0)[2:].strip()

            # Remove leading empty lines again
            while body and not body[0]:
                body.pop(0)

        return meta, body

    def try_load_archetype(
        self, archetypes: Archetypes, relpath: str, name: str
    ) -> MarkdownArchetype | None:
        if os.path.basename(relpath) != name + ".md":
            return None
        return MarkdownArchetype(archetypes, relpath, self)


class MarkdownArchetype(Archetype):
    def __init__(self, archetypes: Archetypes, relpath: str, feature: MarkdownPages):
        super().__init__(archetypes, relpath)
        self.feature = feature

    def render(self, **kw: Any) -> tuple[dict[str, Any], str]:
        meta, rendered = super().render(**kw)

        # Reparse the rendered version
        with io.BytesIO(rendered.encode()) as fd:
            # Reparse it separating front matter and markdown content
            try:
                style, meta, body = front_matter.read_markdown_partial(fd)
            except Exception as e:
                raise RuntimeError(
                    f"archetype {self.relpath}: failed to parse front matter"
                ) from e

            if style is None:
                raise RuntimeError(
                    f"archetype {self.relpath}: unrecognized style of front matter"
                )

            # Make a copy of the full parsed metadata
            archetype_meta = dict(meta)

            # Remove the path entry
            meta.pop("path", None)

            # Reserialize the page with the edited metadata
            fmatter = front_matter.write(meta, style)
            with io.StringIO() as out:
                out.write(fmatter)
                print(file=out)
                for line in body:
                    print(line, file=out)
                post_body = out.getvalue()

        return archetype_meta, post_body


class MarkdownPage(TemplatePage, MarkupPage, FrontMatterPage):
    """
    Markdown sources

    Markdown files have a `.md` extension and are prefixed by
    [front matter metadata](../front-matter.md).

    The flavour of markdown is what's supported by
    [python-markdown](http://pythonhosted.org/Markdown/) with the
    [Extra](http://pythonhosted.org/Markdown/extensions/extra.html),
    [CodeHilite](http://pythonhosted.org/Markdown/extensions/code_hilite.html)
    and [Fenced Code Blocks](http://pythonhosted.org/Markdown/extensions/fenced_code_blocks.html)
    extensions, and is quite close to
    [GitHub Flavored Markdown](https://github.github.com/gfm/) or
    [GitLab Markdown](https://docs.gitlab.com/ee/user/markdown.html).

    `staticsite` adds an extra internal plugin to Python-Markdown to postprocess
    the page contents to adjust internal links to guarantee that they point where
    they should.

    Adding a horizontal rule *using underscores* (3 or more underscores), creates a
    page fold. When rendering the page inline, such as in a blog index page, or in
    RSS/Atom syndication, the content from the horizontal rule onwards will not be
    shown.

    If you want to add a horizontal rule without introducing a page fold, use a
    sequence of three or more asterisks (`***`) or dashes (`---`) instead.


    ## Linking to other pages

    Pages can link to other pages via normal Markdown links (`[text](link)`).

    Links that start with a `/` will be rooted at the top of the site contents.

    Relative links are resolved relative to the location of the current page first,
    and failing that relative to its parent directory, and so on until the root of
    the site.

    For example, if you have `blog/2016/page.md` that contains a link like
    `![a photo](images/photo.jpg)`, the link will point to the first of this
    options that will be found:

    1. `blog/2016/images/photo.jpg`
    2. `blog/images/photo.jpg`
    3. `images/photo.jpg`

    This allows one to organise pages pointing to other pages or assets without needing
    to worry about where they are located in the site.

    You can link to other Markdown pages with the `.md` extension
    ([like GitHub does](https://help.github.com/articles/relative-links-in-readmes/))
    or without, as if you were editing a wiki.

    ## Page metadata

    The front matter of the post can be written in
    [TOML](https://github.com/toml-lang/toml),
    [YAML](https://en.wikipedia.org/wiki/YAML) or
    [JSON](https://en.wikipedia.org/wiki/JSON), just like in
    [Hugo](https://gohugo.io/content/front-matter/).

    Use `---` delimiters to mark YAML front matter. Use `+++` delimiters to mark
    TOML front matter. Use `{`…`}` delimiters to mark JSON front matter.

    You can also usea [triple-backticks code blocks](https://python-markdown.github.io/extensions/fenced_code_blocks/)
    first thing in the file to mark front matter, optionally specifying `yaml`,
    `toml`, or `json` as the format (yaml is used as a default):

    ~~~~{.markdown}
    ```yaml
    date: 2020-01-02 12:00
    ```
    # My page
    ~~~~

    If you want to start your markdown content with a code block, add an empty line
    at the top: front matter detection only happens on the first line of the file.

    See [page metadata](../metadata.md) for a list of commonly used metadata.


    ## Extra settings

    Markdown rendering makes use of these settings:

    ### `MARKDOWN_EXTENSIONS`

    Extensions used by python-markdown. Defaults to:

    ```py
    MARKDOWN_EXTENSIONS = [
        "markdown.extensions.extra",
        "markdown.extensions.codehilite",
        "markdown.extensions.fenced_code",
    ]
    ```

    ### `MARKDOWN_EXTENSION_CONFIGS`

    Configuration for markdown extensions. Defaults to:

    ```py
    MARKDOWN_EXTENSION_CONFIGS = {
        'markdown.extensions.extra': {
            'markdown.extensions.footnotes': {
                # See https://github.com/spanezz/staticsite/issues/13
                'UNIQUE_IDS': True,
            },
        },
    }
    ```

    ## Rendering markdown pages

    Besides the usual `meta`, markdown pages have also these attributes:

    * `page.contents`: the Markdown contents rendered as HTML. You may want to use
      it with the [`|safe` filter](https://jinja.palletsprojects.com/en/2.10.x/templates/#safe)
      to prevent double escaping
    """

    TYPE = "markdown"

    # Match a Markdown divider line
    re_divider = re.compile(r"^____+$")

    def __init__(self, *, body: list[str], **kw: Any):
        self.feature: MarkdownPages
        # Indexed by default
        kw.setdefault("indexed", True)
        super().__init__(**kw)

        # Sequence of lines found in the body before the divider line, if any
        self.body_start: list[str]

        # Sequence of lines found in the body including and after the divider
        # line, nor None if there is no divider line
        self.body_rest: list[str] | None

        # External links found when rendering the page
        self.rendered_external_links: set[str] = set()

        # Split lead and rest of the post, if a divider line is present
        for idx, line in enumerate(body):
            if self.re_divider.match(line):
                self.body_start = body[:idx]
                self.body_rest = body[idx:]
                break
        else:
            self.body_start = body
            self.body_rest = None

    def front_matter_changed(self, fd: BinaryIO) -> bool:
        """
        Check if the front matter read from fd is different from ours
        """
        # TODO: refactor read_file_meta to skip reading the body if we don't
        # need it, but still parse until the title if the front matter doesn't
        # have one
        meta, body = self.feature.read_file_meta(fd)
        return self.front_matter != meta

    def check(self) -> None:
        self.render()

    def _render_page(
        self, body: list[str], render_type: str, absolute: bool = False
    ) -> str:
        """
        Render markdown in the context of the given page.
        """
        cache_key = f"{render_type}:{self.src.relpath}"

        with self.markup_render_context(cache_key, absolute=absolute) as context:
            if rendered := context.cache.get("rendered"):
                # log.info("%s: markdown cache hit", page.src.relpath)
                return cast(str, rendered)

            self.feature.markdown.reset()
            rendered = self.feature.markdown.convert("\n".join(body))

            self.rendered_external_links.update(
                self.feature.link_resolver.external_links
            )

            context.cache["rendered"] = rendered

        return rendered

    @jinja2.pass_context
    def html_body(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        absolute = self != context["page"]
        if self.body_rest is not None:
            body = self.body_start + ["", "<a name='sep'></a>", ""] + self.body_rest
            return self._render_page(body, render_type="hb", absolute=absolute)
        else:
            return self._render_page(
                self.body_start, render_type="s", absolute=absolute
            )

    @jinja2.pass_context
    def html_inline(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        absolute = self != context["page"]
        if self.body_rest is not None:
            body = self.body_start + ["", f"[(continue reading)](/{self.src.relpath})"]
            return self._render_page(body, render_type="h", absolute=absolute)
        else:
            return self._render_page(
                self.body_start, render_type="s", absolute=absolute
            )

    @jinja2.pass_context
    def html_feed(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        absolute = self != context["page"]
        if self.body_rest is not None:
            body = self.body_start + [""] + self.body_rest
            return self._render_page(body, render_type="f", absolute=absolute)
        else:
            return self._render_page(
                self.body_start, render_type="f", absolute=absolute
            )


FEATURES = {
    "md": MarkdownPages,
}



from __future__ import annotations

import functools
import heapq
import logging
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, Union, cast

from staticsite import fields
from staticsite.feature import Feature, TrackedField
from staticsite.features.syndication import Syndication
from staticsite.page import (
    AutoPage,
    ChangeExtent,
    ConstPageField,
    Page,
    SourcePage,
    TemplatePage,
)
from staticsite.utils import front_matter

if TYPE_CHECKING:
    from staticsite import file, fstree
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("taxonomy")


class Taxonomy:
    """
    Definition of a taxonomy for the site
    """

    def __init__(self, *, name: str, src: file.File, **kw: Any):
        self.name = name
        self.src = src
        self.index: TaxonomyPage | None = None
        # Pages that declare categories in this taxonomy
        self.pages: set[Page] = set()
        # Metadata for the taxonomy index page
        self.index_meta: dict[str, Any] = {"taxonomy": self}
        # Metadata for the category pages
        self.category_meta: dict[str, Any] = {}
        # Metadata for the archive pages
        self.archive_meta: dict[str, Any] = {}
        # Category pages by category name
        self.category_pages: dict[str, CategoryPage] = {}

    def update_meta(
        self,
        category: dict[str, Any] | None = None,
        archive: dict[str, Any] | None = None,
        **kw: Any,
    ) -> None:
        """
        Add information from a parsed taxonomy file
        """
        self.index_meta.update(kw)

        # Metadata for category pages
        if category is not None:
            self.category_meta.update(category)
        self.category_meta.setdefault("template_title", "{{page.name}}")

        # Metadata for archive pages
        if archive is not None:
            self.archive_meta.update(archive)

    def create_index(self, node: SourcePageNode) -> TaxonomyPage | None:
        """
        Create the index page for this taxonomy
        """
        self.index = node.create_source_page_as_path(
            page_cls=TaxonomyPage, src=self.src, name=self.name, **self.index_meta
        )
        return self.index

    def create_category_page(self, name: str, pages: list[Page]) -> CategoryPage | None:
        """
        Generate the page for one category in this taxonomy
        """
        if self.index is None:
            raise RuntimeError(
                "Taxonomy.create_category_page called without an index page"
            )
        # Sort pages by date, used by series sequencing
        pages.sort(key=lambda p: p.date)

        # Create category page
        category_meta = dict(self.category_meta)
        category_meta["pages"] = pages
        category_meta["date"] = pages[-1].date

        # Syndication
        if (syndication_value := category_meta.pop("syndication", None)) is not None:
            syndication = Syndication.clean_value(None, syndication_value)
        else:
            syndication = Syndication(None)

        if syndication is not None:
            category_meta["syndication"] = syndication
            category_meta["syndication"].archive.update(self.archive_meta)

            # Don't auto-add feeds for the tag syndication pages
            syndication.add_to = False

        return self.index.node.create_auto_page_as_path(
            created_from=self.index, page_cls=CategoryPage, name=name, **category_meta
        )

    def generate_pages(self) -> None:
        if self.index is None:
            raise RuntimeError("Taxonomy.generate_pages called without an index page")
        self.category_meta["taxonomy"] = self.index

        # Group pages by category
        by_category: dict[str, list[Page]] = defaultdict(list)
        for page in self.pages:
            categories = getattr(page, self.name, None)
            if not categories:
                continue
            # File the page in its category lists
            for category in categories:
                by_category[category].append(page)

        # Create category pages
        for category, pages in by_category.items():
            if (cpage := self.create_category_page(category, pages)) is not None:
                self.category_pages[category] = cpage

        # Replace category names with category pages in each categorized page
        for page in self.pages:
            if not (categories := getattr(page, self.name, None)):
                continue
            setattr(
                page,
                self.name,
                [
                    p
                    for c in categories
                    if (p := self.category_pages.get(c)) is not None
                ],
            )

        # Sort categories dict by category name
        self.category_pages = {k: v for k, v in sorted(self.category_pages.items())}

        # Set self.meta.pages to the sorted list of categories
        self.index.pages = list(self.category_pages.values())


class TaxonomyField(TrackedField[Page, Union[list[str], list[Page]]]):
    tracked_by = "taxonomy"

    def _clean(
        self, page: Page, value: None | str | list[str] | list[Page]
    ) -> list[str] | list[Page]:
        if not value:
            return cast(list[Page], [])
        elif isinstance(value, str):
            return [value]
        elif isinstance(value, list):
            return value
        else:
            raise ValueError(
                f"{value!r} is not a string, a list of strings, or a list of Page objects"
            )


class BaseTaxonomyPageMixin(Page):
    series_title = fields.Str["BaseTaxonomyPageMixin"](
        doc="""
        Series title from this page onwards.

        If this page is part of a series, and it defines `series_title`, then
        the series title will be changed to this, from this page onwards, but
        not for the previous pages
    """
    )


class TaxonomyPageMixin(BaseTaxonomyPageMixin):
    """
    Base class for dynamically generated taxonomy page mixins
    """

    # This is empty to begin with, to serve as a base for TaxonomyMixins
    # dynamically generated when taxonomies are found


class TaxonomyFeature(Feature):
    """
    Tag pages using one or more taxonomies.

    Files with a `.taxonomy` extension represent where that taxonomy will appear in
    the built site.

    For example, `dir/tags.taxonomy` will become `dir/tags/…` in the built site,
    and contain an index of all categories in the taxonomy, and for each category
    an index page, an archive page, and rss and atom feeds.

    **Added in 1.3**: Removed again the need to list in [`TAXONOMIES` settings](../settings.md)
    the taxonomies used in the site.

    **Added in 1.2**: Any taxonomy you use needs to be explicitly listed in
    settings as a `TAXONOMIES` list of taxonomy names. staticsite prints a warning
    if a `.taxonomy` file is found that is not listed in `TAXONOMIES`.

    The `.taxonomy` file will be a yaml, json, or toml file, like in
    [markdown](../pages/markdown.md) front matter.

    The relevant fields in a taxonomy file are:

    * `title`, `description`: used for the taxonomy index page
    * `category`: metadata for the page generated for each category in the taxonomy
    * `archive`: metadata for the page generated for each category archive

    * **changed in 1.2**: `template_tags`: use `template` instead
    * **changed in 1.2**: `template_tag`: use `category/template` instead
    * **changed in 1.2**: `template_archive`: use `archive/template` instead
    * **Removed in 1.2**: `output_dir` is now ignored, and the taxonomy pages will
      be put in a directory with the same name as the file, without extension
    * **Changed in 1.2**: `item_name` in a `.taxonomy` file does not have a special
      meaning anymore, and templates can still find it in the taxonomy page
      metadata
    * **Added in 1.2**: page.meta["tags"] or other taxonomy names gets substituted,
      from a list of strings, to a list of pages for the taxonomy index, which can
      be used to generate links and iterate subpages in templates without the need
      of specialised functions.
    * **Removed in 1.2**: `series_tags` is now ignored: every category can be used
      to build a series

    Example:

    ```yaml
    ---
    # In staticsite, a taxonomy is a group of attributes like categories or tags.
    #
    # Like in Hugo, you can have as many taxonomies as you want. See
    # https://gohugo.io/taxonomies/overview/ for a general introduction to
    # taxonomies.
    #
    # This file describes the taxonomy for "tags". The name of the taxonomy is
    # taken from the file name.
    #
    # The format of the file is the same that is used for the front matter of
    # posts, again same as in Hugo: https://gohugo.io/content/front-matter/

    # Template for rendering the taxonomy index. Here, it's the page that shows the
    # list of tags
    template: taxonomy.html

    category:
       # Template for rendering the page for one tag. Here, it's the page that shows
       # the latest pages tagged with the tag
       template: blog.html
       # Title used in category pages
       template_title: "Latest posts for tag <strong>{{page.name}}</strong>"
       # Description used in category pages
       template_description: "Most recent posts with tag <strong>{{page.name}}</strong>"

       syndication:
         add_to: no
         archive:
           # Template for rendering the archive page for one tag. Here, it's the page that
           # links to all the pages tagged with the tag
           template_title: "Archive of posts for tag <strong>{{page.name}}</strong>"
           template_description: "Archive of all posts with tag <strong>{{page.name}}</strong>"
    ```


    ### Jinja2 templates

    Each taxonomy defines extra `url_for_*` functions. For example, given a *tags*
    taxonomy with *tag* as singular name:

     * `taxonomies()`: list of all taxonomy index pages, ordered by name
     * `taxonomy(name)`: taxonomy index page for the given taxonomy
     * **Removed in 1.2**: `url_for_tags()`: links to the taxonomy index.
     * **Removed in 1.2**: `url_for_tag(tag)`: use `url_for(category)`
     * **Removed in 1.2**: `url_for_tag_archive(tag)`: use `url_for(category.archive)`
     * **Removed in 1.2**: `url_for_tag_rss(tag)`: see [the syndication feature](syndication.md)
     * **Removed in 1.2**: `url_for_tag_atom(tag)`: see [the syndication feature](syndication.md)

    ## Series of posts

    Any category of any taxonomy collects an ordered list of posts, that can be
    used to generate interlinked series of pages.

    All posts in a category are ordered by date (see [page metadata](markdown.md)).

    Given a [category page](taxonomies.md), the `.sequence(page)` method locates
    the page in the sequence of pages in that category, returning a dictionary
    with these values:

    * `index`: position of this page in the series (starts from 1)
    * `length`: number of pages in the series
    * `first`: first page in the series
    * `last`: last page in the series
    * `prev`: previous page in the series, if available
    * `next`: next page in the series, if available
    * `title`: title for the series

    This example template renders the position of the page in the series, choosing
    as a series the first category of the page in a taxonomy called `series`:

    ```jinja2
    {% with place = page.meta.series[0].sequence(page) %}
    {{place.title}} {{place.index}}/{{place.length}}
    {% endwith %}
    ```

    ### Series title

    The series title is, by default, the title of the first page in the series.

    If a page defines a `series_title` header, then it becomes the series title
    from that page onwards. It is possible to redefine the series title during a
    series, like for example a "My trip" series that later becomes "My trip:
    Italy", "My trip: France", "My trip: Spain", and so on.

    ## Multiple series for a page

    A page is part of one series for each category of each taxonomy it has.
    Templates need to choose which categories are relevant for use in generating
    series navigation links.
    """

    RUN_BEFORE = ["syndication"]

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        # All TaxonomyPages found
        self.taxonomies: dict[str, Taxonomy] = {}
        self.page_bases: list[type[Page]] = []

        self.j2_globals["taxonomies"] = self.jinja2_taxonomies
        self.j2_globals["taxonomy"] = self.jinja2_taxonomy

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return self.page_bases

    def get_used_page_types(self) -> list[type[Page]]:
        return [TaxonomyPage, CategoryPage]

    def track_field(self, field: fields.Field[Any, Any], obj: Page, value: Any) -> None:
        self.taxonomies[field.name].pages.add(obj)

    def register_taxonomy(self, name: str, src: file.File) -> None:
        # Note that if we want to make the tags inheritable, we need to
        # interface with 'rst' (or 'rst' to interface with us) because rst
        # needs to know which metadata items are taxonomies in order to parse
        # them.
        # Instead of making tags inheritable from normal metadata, we can offer
        # them to be added by 'files' or 'dirs' directives.
        self.page_bases.append(
            type(
                "TaxonomyMixin",
                (TaxonomyPageMixin,),
                {
                    name: TaxonomyField(
                        structure=True,
                        default=(),
                        doc=f"""
                List of categories for the `{name}` taxonomy.

                Setting this as a simple string is the same as setting it as a list of one
                element.
            """,
                    )
                },
            )
        )
        self.taxonomies[name] = Taxonomy(name=name, src=src)

    def load_dir_meta(self, directory: fstree.Tree) -> dict[str, Any] | None:
        for fname, src in directory.files.items():
            if not fname.endswith(".taxonomy"):
                continue
            self.register_taxonomy(fname[:-9], src)
        return None

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            if not fname.endswith(".taxonomy"):
                continue
            taken.append(fname)

            taxonomy = self.taxonomies[fname[:-9]]

            try:
                fm_meta = self.load_file_meta(directory, fname)
            except Exception:
                log.exception("%s: cannot parse taxonomy information", src.relpath)
                continue
            kwargs.update(fm_meta)

            taxonomy.update_meta(**kwargs)
            if (page := taxonomy.create_index(node)) is None:
                continue
            pages.append(page)

        for fname in taken:
            del files[fname]

        return pages

    def load_file_meta(self, directory: fstree.Tree, fname: str) -> dict[str, Any]:
        """
        Parse the taxonomy file to read its description
        """
        with directory.open(fname, "rt") as fd:
            fmt, meta = front_matter.read_whole(fd)
        return meta

    def jinja2_taxonomies(self) -> Iterable[TaxonomyPage]:
        return [t.index for t in self.taxonomies.values() if t.index is not None]

    def jinja2_taxonomy(self, name: str) -> TaxonomyPage | None:
        if taxonomy := self.taxonomies.get(name):
            return taxonomy.index
        return None

    def generate(self) -> None:
        # Call analyze on all taxonomy pages, to populate them by scanning
        # site pages
        for taxonomy in self.taxonomies.values():
            taxonomy.generate_pages()


class TaxonomyPage(TemplatePage, SourcePage):
    """
    Root page for one taxonomy defined in the site

    Pages generated by the taxonomy feature have these extra attributes:

    * `name`: the taxonomy name (e.g. "tags")
    * `categories`: dict mapping category names to category index pages. In
      templates, the dict is sorted by category name.
    * The page can be indexed by category name, returning the corresponding
      category index page.
    """

    TYPE = "taxonomy"
    TEMPLATE = "taxonomy.html"

    taxonomy = fields.ConstTypeField["TaxonomyPage", Taxonomy](
        cls=Taxonomy, doc="Structured taxonomy information"
    )

    def __init__(self, *args: Any, node: SourcePageNode, **kw: Any):
        kw.setdefault("nav_title", node.name.capitalize())
        super().__init__(*args, node=node, **kw)

        # Taxonomy name (e.g. "tags")
        self.name = self.node.name

    @property
    def categories(self) -> dict[str, CategoryPage]:
        """
        Map all possible values for this taxonomy to the pages that reference
        them
        """
        return self.taxonomy.category_pages

    def to_dict(self) -> dict[str, Any]:
        from staticsite.utils import dump_meta

        res = super().to_dict()
        res["name"] = self.name
        res["categories"] = dump_meta(self.categories)
        return res

    def __getitem__(self, name: str) -> CategoryPage:
        return self.categories[name]

    def top_categories(self, count: int = 10) -> list[CategoryPage]:
        """
        Return the ``count`` categories with the most pages
        """
        return heapq.nlargest(
            count,
            self.categories.values(),
            key=lambda c: len(c.pages) if c.pages else 0,
        )

    def most_recent(self, count: int = 10) -> list[CategoryPage]:
        """
        Return the ``count`` categories with the most recent date
        """
        return heapq.nlargest(count, self.categories.values(), key=lambda c: c.date)

    def _compute_footprint(self) -> dict[str, Any]:
        res = super()._compute_footprint()
        # Map categories to list of page relpaths
        taxonomy: dict[str, list[str]] = {}
        for name, category in self.taxonomy.category_pages.items():
            if not category.pages:
                taxonomy[name] = []
            else:
                taxonomy[name] = [
                    page.src.relpath
                    for page in category.pages
                    if isinstance(page, SourcePage)
                ]
        res["taxonomy"] = taxonomy
        return res

    def _compute_change_extent(self) -> ChangeExtent:
        if (old := self.footprint.get("taxonomy")) is None:
            return ChangeExtent.ALL

        # Check if categories have been added or removed
        if old.keys() != self.taxonomy.category_pages.keys():
            return ChangeExtent.ALL

        # Aggregate change extends from category pages
        if self.taxonomy.category_pages:
            return max(
                page.change_extent for page in self.taxonomy.category_pages.values()
            )
        else:
            return ChangeExtent.UNCHANGED


@functools.total_ordering
class CategoryPage(TemplatePage, AutoPage):
    """
    Index page showing all the pages tagged with a given taxonomy item

    Pages generated by the taxonomy feature have these extra attributes:

    * `page.name`: the category name
    * `page.meta.pages`: the pages that have this category
    * `page.meta.taxonomy`: the taxonomy index page for this category
    * `page.meta.archive`: the archive page for this category

    **Category archive**:

    This is created automatically by the [syndication](../features/syndication.md) feature.

    You can refer to the category page from its archive page, using
    `page.created_from`.
    """

    TYPE = "category"
    TEMPLATE = "blog.html"

    taxonomy = ConstPageField["CategoryPage", TaxonomyPage](
        cls=TaxonomyPage, doc="Page that defined this taxonomy"
    )
    name = fields.Str["CategoryPage"](doc="Name of the category shown in this page")

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        # Category name
        self.name = self.node.name
        # Index of each page in the category sequence
        self.page_index: dict[Page, int] = (
            {page: idx for idx, page in enumerate(self.pages)} if self.pages else {}
        )

    def to_dict(self) -> dict[str, Any]:
        res = super().to_dict()
        res["name"] = self.name
        return res

    def __lt__(self, o: Any) -> bool:
        o_taxonomy = getattr(o, "taxonomy", None)
        if o_taxonomy is None:
            return NotImplemented

        o_name = getattr(o, "name", None)
        if o_name is None:
            return NotImplemented

        return (self.taxonomy.name, self.name) < (o_taxonomy.name, o_name)

    def __eq__(self, o: Any) -> bool:
        o_taxonomy = getattr(o, "taxonomy", None)
        if o_taxonomy is None:
            return NotImplemented

        o_name = getattr(o, "name", None)
        if o_name is None:
            return NotImplemented

        return (self.taxonomy.name, self.name) == (o_taxonomy.name, o_name)

    def __hash__(self) -> int:
        return hash((self.taxonomy.name, self.name))

    @functools.cached_property
    def series_info(self) -> dict[str, Any] | None:
        """
        Return a dict describing this category as a series
        """
        # Compute a series title for this page.
        # Look for the last defined series title, defaulting to the title of
        # the first page in the series.
        if (pages := self.pages) is None:
            return None
        series_title = (
            cast(BaseTaxonomyPageMixin, pages[0]).series_title or pages[0].title
        )
        return {
            # Array with all the pages in the series
            "pages": pages,
            "length": len(pages),
            "first": pages[0],
            "last": pages[-1],
            "title": series_title,
        }

    def sequence(self, page: Page) -> dict[str, Any] | None:
        idx = self.page_index.get(page)
        if idx is None:
            return None

        # Compute a series title for this page.
        # Look for the last defined series title, defaulting to the title of
        # the first page in the series.
        if (pages := self.pages) is None:
            return None
        series_title = pages[0].title
        for p in pages:
            if (title := cast(BaseTaxonomyPageMixin, p).series_title) is not None:
                series_title = title
            if p == page:
                break

        return {
            # Array with all the pages in the series
            "pages": pages,
            # Assign series_prev and series_next metadata elements to pages
            "index": idx + 1,
            "length": len(pages),
            "first": pages[0],
            "last": pages[-1],
            "prev": pages[idx - 1] if idx > 0 else None,
            "next": pages[idx + 1] if idx < len(pages) - 1 else None,
            "title": series_title,
        }

    def _compute_change_extent(self) -> ChangeExtent:
        if self.created_from is None:
            raise AssertionError(f"CategoryPage {self!r} found with empty created_from")

        index = cast(TaxonomyPage, self.created_from)

        # No previous footprint
        if index.old_footprint is None:
            return ChangeExtent.ALL
        if (old_footprint := index.old_footprint.get("taxonomy")) is None:
            return ChangeExtent.ALL

        # This category did not previously exist
        if (old_sources_list := old_footprint.get(self.name)) is None:
            return ChangeExtent.ALL
        else:
            old_sources = set(old_sources_list)

        sources = (
            {page.src.relpath for page in self.pages if isinstance(page, SourcePage)}
            if self.pages
            else {}
        )
        if sources != old_sources:
            # If any page added or removed this category, we rebuild
            return ChangeExtent.ALL

        if self.pages:
            return max(p.change_extent for p in self.pages)
        else:
            return ChangeExtent.ALL


class TestTaxonomyPage(TaxonomyPage):
    def _read_taxonomy_description(self) -> None:
        pass


FEATURES = {
    "taxonomy": TaxonomyFeature,
}



from __future__ import annotations

import logging
from collections.abc import Iterator, Sequence
from typing import Any, cast

from staticsite import fields
from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.node import Node
from staticsite.page import Page, PageNotFoundError

log = logging.getLogger("nav")


class NavData:
    def __init__(self, page: Page, paths: list[str]):
        self.page = page
        self.paths = paths
        self.resolved: list[NavPageMixin] | None = None

    def _resolve(self) -> None:
        if self.resolved is not None:
            return
        self.resolved = []
        for path in self.paths:
            try:
                page = self.page.resolve_path(path)
            except PageNotFoundError as e:
                log.warning("%r: skipping page in nav: %s", self.page, e)
                continue
            self.resolved.append(cast(NavPageMixin, page))

    def __iter__(self) -> Iterator[NavPageMixin]:
        self._resolve()
        return cast(list["NavPageMixin"], self.resolved).__iter__()

    def to_dict(self) -> list[NavPageMixin]:
        self._resolve()
        return cast(list["NavPageMixin"], self.resolved)


class NavField(TrackedField[Page, NavData]):
    """
    List of page paths, relative to the page defining the nav element, that
    are used for the navbar.
    """

    tracked_by = "nav"

    def __init__(self, **kw: Any):
        kw.setdefault("inherited", True)
        kw.setdefault("default", ())
        super().__init__(**kw)

    def _clean(self, page: Page, value: Any) -> NavData:
        """
        Set metadata values in obj from values
        """
        return NavData(page, value)


class NavNodeMixin(Node):
    nav = NavField(structure=True)


class NavPageMixin(Page):
    nav = NavField(structure=True)
    nav_title = fields.Str[Page](
        doc="""
        Title to use when this page is linked in a navbar.

        It defaults to `page.title`, or to the series name for series pages.

        `nav_title` is only guaranteed to exist for pages that are used in `nav`.
    """
    )


class Nav(PageTrackingMixin[NavPageMixin], Feature):
    """
    Expand a 'pages' metadata containing a page filter into a list of pages.
    """

    def get_node_bases(self) -> Sequence[type[Node]]:
        return (NavNodeMixin,)

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return (NavPageMixin,)

    def crossreference(self) -> None:
        # Expand pages expressions
        nav_pages: set[NavPageMixin] = set()

        for page in self.tracked_pages:
            if (nav := page.nav) is None:
                continue

            # Resolve paths to target pages
            nav._resolve()

            # Build list of target pages
            if nav.resolved is not None:
                nav_pages.update(nav.resolved)

        # Make sure nav_title is filled
        for target in nav_pages:
            if not target.nav_title:
                target.nav_title = target.title


FEATURES = {
    "nav": Nav,
}



from __future__ import annotations

import io
import logging
import os
from typing import IO, TYPE_CHECKING, Any, cast

import docutils.core
import docutils.io
import docutils.nodes
import docutils.writers.html5_polyglot
import jinja2

from staticsite.archetypes import Archetype
from staticsite.feature import Feature
from staticsite.markup import MarkupFeature, MarkupPage
from staticsite.page import FrontMatterPage, Page, TemplatePage
from staticsite.utils import yaml_codec

if TYPE_CHECKING:
    import docutils.node

    from staticsite import file, fstree
    from staticsite.archetypes import Archetypes
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("rst")


class DoctreeScan:
    def __init__(self, doctree: docutils.nodes.document) -> None:
        # Doctree root node
        self.doctree = doctree
        # Information useful to locate and remove the docinfo element
        self.docinfo: docutils.nodes.docinfo | None = None
        self.docinfo_idx: int | None = None
        # First title element
        self.first_title: docutils.nodes.title | None = None
        # All <target> link elements that need rewriting on rendering
        self.links_target: list[docutils.nodes.target | docutils.nodes.reference] = []
        # All <image> link elements that need rewriting on rendering
        self.links_image: list[docutils.nodes.image] = []
        # Return True if the link targets have been rewritten
        self.links_rewritten = False

        # Scan tree contents looking for significant nodes
        self.scan(self.doctree)

    def remove_docinfo(self) -> None:
        # Remove docinfo element from tree
        if self.docinfo is not None:
            self.docinfo.parent.pop(self.docinfo_idx)

    def scan(self, element: docutils.node.Node) -> None:
        """
        Scan the doctree collecting significant elements
        """
        for idx, node in enumerate(element.children):
            if self.first_title is None and isinstance(node, docutils.nodes.title):
                self.first_title = node
            elif isinstance(node, (docutils.nodes.target, docutils.nodes.reference)):
                self.links_target.append(node)
            elif isinstance(node, docutils.nodes.image):
                self.links_image.append(node)
            elif isinstance(node, docutils.nodes.docinfo):
                self.docinfo = node
                self.docinfo_idx = idx
                # Do not recurse into a docinfo
                continue

            # Recursively descend
            self.scan(node)


class RestructuredText(MarkupFeature, Feature):
    """
    Render ``.rst`` reStructuredText pages, with front matter.

    See doc/reference/rst.rst for details.
    """

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)

        self.render_cache = self.site.caches.get("rst")

        # Names of tags whose content should be parsed as yaml
        self.yaml_tags = {"files", "dirs"}
        self.yaml_tags_filled = False

    def get_used_page_types(self) -> list[type[Page]]:
        return [RstPage]

    def parse_rest(
        self, fd: IO[str], remove_docinfo: bool = True
    ) -> tuple[dict[str, Any], DoctreeScan]:
        """
        Parse a rest document.

        Return a tuple of 2 elements:
            * a dict with the first docinfo entries
            * the doctree with the docinfo removed
        """
        # Parse input into doctree
        doctree = docutils.core.publish_doctree(fd, source_class=docutils.io.FileInput)

        doctree_scan = DoctreeScan(doctree)

        # Get metadata fields from docinfo
        meta = {}
        if doctree_scan.docinfo is not None:
            for child in doctree_scan.docinfo.children:
                if child.tagname == "field":
                    name = child.attributes.get("classes")[0]
                    for fchild in child.children:
                        if fchild.tagname == "field_body":
                            meta[name] = fchild.astext().strip()
                else:
                    meta[child.tagname] = child.astext().strip()

        if doctree_scan.first_title is not None:
            if "title" not in meta:
                meta["title"] = doctree_scan.first_title.astext()
            # If the title element is at the beginning of the doctree, remove
            # it to avoid a duplicate title in the rendered content
            if (
                doctree_scan.docinfo
                and doctree_scan.docinfo.children
                and doctree_scan.docinfo.children[0] == doctree_scan.first_title
            ):
                doctree_scan.doctree.children.pop(0)

        # If requested, parse some tag contents as yaml
        if self.yaml_tags:
            for tag in self.yaml_tags:
                val = meta.get(tag)
                if val is not None and isinstance(val, str):
                    meta[tag] = yaml_codec.loads(val)

        if remove_docinfo:
            doctree_scan.remove_docinfo()

        return meta, doctree_scan

    def load_dir_meta(self, directory: fstree.Tree) -> dict[str, Any] | None:
        # Load front matter from index.rst
        # Do not try to load front matter from README.md, as one wouldn't
        # clutter a repo README with staticsite front matter
        if "index.rst" not in directory.files:
            return None

        # Parse to get at the front matter
        with directory.open("index.rst", "rt") as fd:
            meta, doctree_scan = self.parse_rest(fd, remove_docinfo=False)

        return meta

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        if not self.yaml_tags_filled:
            cls = self.site.features.get_page_class(RstPage)
            for name, field in cls._fields.items():
                if field.structure:
                    self.yaml_tags.add(name)
            self.yaml_tags_filled = True

        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            if not fname.endswith(".rst"):
                continue
            taken.append(fname)

            try:
                fm_meta, doctree_scan = self.load_file_meta(directory, fname)
            except Exception as e:
                log.debug(
                    "%s: Failed to parse RestructuredText page: skipped",
                    src,
                    exc_info=True,
                )
                log.warning(
                    "%s: Failed to parse RestructuredText page: skipped (%s)", src, e
                )
                continue

            kwargs.update(fm_meta)
            kwargs["page_cls"] = RstPage
            kwargs["src"] = src
            kwargs["feature"] = self
            kwargs["front_matter"] = fm_meta
            kwargs["doctree_scan"] = doctree_scan

            if fname in ("index.rst", "README.rst"):
                page = node.create_source_page_as_index(**kwargs)
            else:
                page = node.create_source_page_as_path(name=fname[:-4], **kwargs)
            if page is not None:
                pages.append(page)

        for fname in taken:
            del files[fname]

        return pages

    def load_file_meta(
        self, directory: fstree.Tree, fname: str
    ) -> tuple[dict[str, Any], DoctreeScan]:
        # Parse document into a doctree and extract docinfo metadata
        with directory.open(fname, "rt") as fd:
            meta, doctree_scan = self.parse_rest(fd)

        return meta, doctree_scan

    def try_load_archetype(
        self, archetypes: Archetypes, relpath: str, name: str
    ) -> Archetype | None:
        if os.path.basename(relpath) != name + ".rst":
            return None
        return RestArchetype(archetypes, relpath, self)


class RestArchetype(Archetype):
    def __init__(self, archetypes: Archetypes, relpath: str, feature: RestructuredText):
        super().__init__(archetypes, relpath)
        self.feature = feature

    def render(self, **kw: Any) -> tuple[dict[str, Any], str]:
        meta, rendered = super().render(**kw)

        # Reparse the rendered version
        with io.StringIO(rendered) as fd:
            parsed_meta, doctree_scan = self.feature.parse_rest(
                fd, remove_docinfo=False
            )

        meta.update(**parsed_meta)

        # # Remove path from docinfo
        # if doctree_scan.docinfo:
        #     for idx, node in doctree_scan.docinfo.children:
        #         if node.tagname == 'field' and node.attributes.get('classes')[0] == "path":
        #             doctree_scan.docinfo.pop(idx)
        #             break
        #
        # It would be good if we could now serialize the doctree back to reSt,
        # but there seems to be no writer for that.
        #
        # But no. Fall back to brutally hacking the rendered text.
        if doctree_scan.docinfo:
            lines = []
            head = True
            for line in rendered.splitlines():
                line = line.rstrip()
                if head:
                    if not line:
                        lines.append(line)
                        head = False
                    else:
                        if not line.startswith(":path: "):
                            lines.append(line)
                else:
                    lines.append(line)
            rendered = "\n".join(lines) + "\n"

        return meta, rendered


class RstPage(FrontMatterPage, MarkupPage, TemplatePage):
    """
    RestructuredText files

    RestructuredText files have a `.rst` extension and their metadata are taken
    from docinfo information.

    `staticsite` will postprocess the RestructuredText doctree to adjust internal
    links to guarantee that they point where they should.


    ## Linking to other pages

    Pages can link to other pages via any of the normal reSt links.

    Links that start with a `/` will be rooted at the top of the site contents.

    Relative links are resolved relative to the location of the current page first,
    and failing that relative to its parent directory, and so on until the root of
    the site.

    For example, if you have `blog/2016/page.rst` that contains a link to
    `images/photo.jpg`, the link will point to the first of this
    options that will be found:

    1. `blog/2016/images/photo.jpg`
    2. `blog/images/photo.jpg`
    3. `images/photo.jpg`

    This allows organising pages pointing to other pages or assets without needing
    to worry about where they are located in the site.

    You can link to other Markdown or RestructuredText pages with the `.md` or
    `.rst` extension ([like GitHub does](https://help.github.com/articles/relative-links-in-readmes/))
    or without, as if you were editing a wiki.


    Page metadata
    -------------

    As in [Sphinx](http://www.sphinx-doc.org/en/stable/markup/misc.html#file-wide-metadata),
    a field list near the top of the file is parsed as front matter and removed
    from the generated files.

    All [bibliographic fields](http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#bibliographic-fields)
    known to docutils are parsed according to their respective type.

    All fields whose name matches a taxonomy defined in `TAXONOMY_NAMES`
    [settings](../settings.md)_ are parsed as yaml, and expected to be a list of
    strings, with the set of values (e.g. tags) of the given taxonomy for the
    current page.

    See [page metadata](../metadata.md) for a list of commonly used metadata.


    ## Rendering reStructuredText pages
    --------------------------------

    Besides the usual `meta`, reStructuredText pages have also these attributes:

    * `page.contents`: the reSt contents rendered as HTML. You may want to use
      it with the [`|safe` filter](https://jinja.palletsprojects.com/en/2.10.x/templates/#safe)
      to prevent double escaping
    """

    TYPE = "rst"

    def __init__(self, *, doctree_scan: DoctreeScan, **kw: Any):
        self.feature: RestructuredText
        # Indexed by default
        kw.setdefault("indexed", True)
        super().__init__(**kw)

        # Document doctree root node
        self.doctree_scan = doctree_scan

    def front_matter_changed(self, fd: IO[str]) -> bool:
        """
        Check if the front matter read from fd is different from ours
        """
        meta, doctree_scan = self.feature.parse_rest(fd)
        return self.front_matter != meta

    def check(self) -> None:
        self._render_page()

    def _render_page(self, absolute: bool = False) -> str:
        cache_key = self.src.relpath
        with self.markup_render_context(cache_key, absolute=absolute) as context:
            if cached := context.cache.get("rendered"):
                # log.info("%s: rst cache hit", page.src.relpath)
                return cast(str, cached)

            if not self.doctree_scan.links_rewritten:
                for node in self.doctree_scan.links_target:
                    node.attributes["refuri"] = context.link_resolver.resolve_url(
                        node.attributes["refuri"]
                    )
                for node in self.doctree_scan.links_image:
                    node.attributes["uri"] = context.link_resolver.resolve_url(
                        node.attributes["uri"]
                    )
                self.doctree_scan.links_rewritten = True

            writer = docutils.writers.html5_polyglot.Writer()
            # TODO: study if/how we can con configure publish_programmatically to
            # do as little work as possible
            output, pub = docutils.core.publish_programmatically(
                source=self.doctree_scan.doctree,
                source_path=None,
                source_class=docutils.io.DocTreeInput,
                destination=None,
                destination_path=None,
                destination_class=docutils.io.StringOutput,
                reader=None,
                reader_name="doctree",
                parser=None,
                parser_name="null",
                writer=writer,
                writer_name=None,
                settings=None,
                settings_spec=None,
                settings_overrides=None,
                config_section=None,
                enable_exit_status=False,
            )
            parts = pub.writer.parts
            rendered: str = parts["body"]
            context.cache["rendered"] = rendered
            return rendered

    @jinja2.pass_context
    def html_body(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        return self._render_page(absolute=self != context["page"])

    @jinja2.pass_context
    def html_inline(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        return self._render_page(absolute=self != context["page"])

    @jinja2.pass_context
    def html_feed(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        return self._render_page(absolute=self != context["page"])


FEATURES = {
    "rst": RestructuredText,
}



from __future__ import annotations

import json
import os
from functools import cached_property
from typing import Any, Protocol

try:
    import lmdb

    HAVE_LMDB = True
except ModuleNotFoundError:
    HAVE_LMDB = False


class Cache(Protocol):
    def __init__(self, fname: str):
        ...

    def get(self, relpath: str) -> Any:
        ...

    def put(self, relpath: str, data: Any) -> None:
        ...


CacheImplementation: type[Cache]


if HAVE_LMDB:

    class LMDBCache:
        def __init__(self, fname: str):
            self.fname = fname + ".lmdb"

        @cached_property
        def db(self) -> lmdb.Environment:
            os.makedirs(os.path.dirname(self.fname), exist_ok=True)
            return lmdb.open(
                self.fname, metasync=False, sync=False, map_size=100 * 1024 * 1024
            )

        def get(self, relpath: str) -> Any:
            with self.db.begin() as tr:
                res = tr.get(relpath.encode(), None)
                if res is None:
                    return None
                else:
                    return json.loads(res)

        def put(self, relpath: str, data: Any) -> None:
            with self.db.begin(write=True) as tr:
                tr.put(relpath.encode(), json.dumps(data).encode())

    CacheImplementation = LMDBCache

else:
    import dbm

    class DBMCache:
        def __init__(self, fname: str):
            self.fname = fname

        # FIXME: using Any here because the dbm module is not typed
        @cached_property
        def db(self) -> Any:
            os.makedirs(os.path.dirname(self.fname), exist_ok=True)
            return dbm.open(self.fname, "c")

        def get(self, relpath: str) -> Any:
            res = self.db.get(relpath)
            if res is None:
                return None
            else:
                return json.loads(res)

        def put(self, relpath: str, data: Any) -> None:
            self.db[relpath] = json.dumps(data)

    CacheImplementation = DBMCache


class DisabledCache:
    """
    noop render cache, for when caching is disabled
    """

    def __init__(self, fname: str):
        self.fname = fname

    def get(self, relpath: str) -> Any:
        return None

    def put(self, relpath: str, data: Any) -> None:
        pass


class Caches:
    """
    Repository of caches that, if kept across site builds, can speed up further
    builds
    """

    def __init__(self, root: str):
        self.root = root

    def get(self, name: str) -> Cache:
        return CacheImplementation(os.path.join(self.root, name))


class DisabledCaches:
    def get(self, name: str) -> Cache:
        return DisabledCache(name)



from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

import jinja2

if TYPE_CHECKING:
    import datetime

    from .site import Site

log = logging.getLogger("fields")

P = TypeVar("P", bound="FieldContainer")
V = TypeVar("V")


class Field(Generic[P, V]):
    """
    Declarative description of a metadata element used by staticsite
    """

    def __init__(
        self,
        *,
        default: V | None = None,
        structure: bool = False,
        internal: bool = False,
        inherited: bool = False,
        doc: str | None = None,
    ):
        """
        :arg name: name of this metadata element
        :arg default: default value when this field has not been set
        :arg structure: set to True if this element is a structured value. Set
                        to false if it is a simple value like an integer or
                        string
        :arg internal: field is for internal use only, and not exported to
                       pages
        :arg doc: documentation for this metadata element
        """
        self.name: str
        self.default: V | None = default
        self.structure: bool = structure
        self.internal: bool = internal
        self.inherited: bool = inherited
        if doc is not None:
            self.__doc__ = doc

    def get_notes(self) -> Iterable[str]:
        if self.inherited:
            yield "Inherited from parent pages"

    def __set_name__(self, owner: type[P], name: str) -> None:
        self.name = name

    def __get__(self, obj: P, type: type[P] | None = None) -> V | None:
        if self.inherited:
            if self.name not in obj.__dict__:
                if obj._parent is not None and self.name in obj._parent._fields:
                    value = getattr(obj._parent, self.name)
                else:
                    value = self.default
                obj.__dict__[self.name] = value
                return cast(V, value)
            else:
                return cast(V, obj.__dict__[self.name])
        else:
            return obj.__dict__.get(self.name, self.default)

    def __set__(self, obj: P, value: Any) -> None:
        obj.__dict__[self.name] = self._clean(obj, value)

    def _clean(self, obj: P, value: Any) -> V:
        """
        Hook to allow to clean values before set
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}._clean for {obj!r}.{self.name} not implemented"
        )


class Const(Field[P, V]):
    """
    Field that takes a const value once and never changes it
    """

    def __get__(self, obj: P, type: type[P] | None = None) -> V:
        if self.name not in obj.__dict__:
            raise RuntimeError(f"{obj!r}.{self.name} has not been set")
        return cast(V, obj.__dict__[self.name])

    def __set__(self, obj: P, value: Any) -> None:
        if self.name in obj.__dict__:
            raise RuntimeError(f"{obj!r}.{self.name} has already been set")
        obj.__dict__[self.name] = self._clean(obj, value)


class ConstTypeField(Const[P, V]):
    """
    Field containing a string
    """

    def __init__(self, *, cls: type[V], **kw: Any):
        super().__init__(**kw)
        self.cls = cls

    def _clean(self, page: P, value: Any) -> V:
        if not isinstance(value, self.cls):
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                f" expecting {self.cls.__name__}"
            )
        return value


class Template(Field[P, jinja2.Template]):
    """
    This metadata, when present in a directory index, should be inherited by
    other files in directories and subdirectories.
    """

    def _clean(self, obj: P, value: Any) -> jinja2.Template:
        # Make sure the template is compiled
        if isinstance(value, jinja2.Template):
            return value
        elif isinstance(value, str):
            return obj.site.theme.jinja2.from_string(value)
        else:
            raise ValueError(f"{value!r} is not a valid value for a template field")


class Str(Field[P, str]):
    """
    Field containing a string
    """

    def _clean(self, obj: P, value: Any) -> str:
        return str(value)


class Int(Field[P, int]):
    """
    Field containing an integer
    """

    def _clean(self, obj: P, value: Any) -> int:
        return int(value)


class ConstInt(Const[P, int]):
    """
    Const field containing an integer
    """

    def _clean(self, obj: P, value: Any) -> int:
        return int(value)


class Float(Field[P, float]):
    """
    Field containing a date
    """

    def _clean(self, obj: P, value: Any) -> float:
        return float(value)


class Date(Field[P, "datetime.datetime"]):
    """
    Field containing a date
    """

    def _clean(self, obj: P, value: Any) -> datetime.datetime:
        return obj.site.clean_date(value)


class Bool(Field[P, bool]):
    """
    Make sure the field is a bool, possibly with a default value
    """

    def __init__(self, *, default: bool | None = None, **kw: Any):
        super().__init__(**kw)
        self.default = default

    def _clean(self, obj: P, val: Any) -> bool:
        if isinstance(val, bool):
            return val
        elif isinstance(val, str):
            return val.lower() in ("yes", "true", "1")
        else:
            raise ValueError(f"{val!r} is not a valid value for a bool field")


class Dict(Field[P, dict[str, Any]]):
    """
    Make sure the field is a dict
    """

    def __get__(self, obj: P, type: type[P] | None = None) -> dict[str, Any]:
        if (value := obj.__dict__.get(self.name)) is None:
            value = {}
            obj.__dict__[self.name] = value
        return value

    def _clean(self, obj: P, value: Any) -> dict[str, Any]:
        """
        Hook to allow to clean values before set
        """
        if isinstance(value, dict):
            return value
        raise ValueError(f"{value!r} is not a valid value for a dict field")


class FieldsMetaclass(type):
    """
    Allow a class to have a set of Field members, defining self-documenting
    metadata elements
    """

    _fields: dict[str, Field[Any, Any]]

    def __new__(
        cls: type[FieldsMetaclass], name: str, bases: tuple[type], dct: dict[str, Any]
    ) -> FieldsMetaclass:
        _fields = {}

        # Add fields from subclasses
        for b in bases:
            if b_fields := getattr(b, "_fields", None):
                _fields.update(b_fields)

        # Add fields from the class itself
        for field_name, val in dct.items():
            if isinstance(val, Field):
                # Store its description in the Model _meta
                _fields[field_name] = val
            else:
                # Leave untouched
                continue

        res = super().__new__(cls, name, bases, dct)
        res._fields = _fields
        return res


class FieldContainer(metaclass=FieldsMetaclass):
    _fields: dict[str, Field[Any, Any]]

    def __init__(self, site: Site, *, parent: FieldContainer | None = None, **kw: Any):
        # Reference to Site, so that fields can access configuration, template
        # compilers, and so on
        self.site = site
        # Reference to parent, for field inheritance
        self._parent = parent

        self.update_fields(kw)

    def update_fields(self, values: dict[str, Any]) -> None:
        for name, value in values.items():
            if name in self._fields:
                setattr(self, name, value)
            else:
                log.warning("%r: setting unknown field %s=%r", self, name, value)



from __future__ import annotations

from unittest import TestCase

from . import utils as test_utils

BASE_FILES = {
    "blog/post1.rst": """
:date: 2016-04-16 10:23:00+02:00

Example blog post in reStructuredText
=====================================
""",
    "blog/post2.md": """---
date: 2016-04-17 10:00
---
# Post 2
""",
    "blog/widget.md": {"syndicated": False},
}


class TestSyndication(test_utils.MockSiteTestMixin, TestCase):
    def test_simple(self):
        files = dict(BASE_FILES)
        files[
            "blog.md"
        ] = """---
date: 2016-04-16 10:23:00+02:00
pages: blog/*
syndication: yes
---

# Title

text
"""
        with self.site(files) as mocksite:
            blog, post1, post2, widget, rss, atom = mocksite.page(
                "blog",
                "blog/post1",
                "blog/post2",
                "blog/widget",
                "blog/index.rss",
                "blog/index.atom",
            )

            synd = blog.syndication
            self.assertEqual(synd.pages, [post2, post1])

            self.assertIsNone(post1.syndication)
            self.assertEqual(
                post1.related,
                {
                    "rss_feed": rss,
                    "atom_feed": atom,
                },
            )

            self.assertIsNone(post2.syndication)
            self.assertEqual(
                post2.related,
                {
                    "rss_feed": rss,
                    "atom_feed": atom,
                },
            )

            self.assertEqual(rss.pages, synd.pages)
            self.assertEqual(rss.title, blog.title)

            self.assertEqual(atom.pages, synd.pages)
            self.assertEqual(atom.title, blog.title)

            self.assertIsNone(widget.syndication)

    def test_add_to_false(self):
        files = dict(BASE_FILES)
        files[
            "blog.md"
        ] = """---
date: 2016-04-16 10:23:00+02:00
pages: blog/*
syndication:
  title: Syndication
  add_to: no
---

# Title

text
"""
        with self.site(files) as mocksite:
            blog, post1, post2, widget, rss, atom = mocksite.page(
                "blog",
                "blog/post1",
                "blog/post2",
                "blog/widget",
                "blog/index.rss",
                "blog/index.atom",
            )

            synd = blog.syndication
            self.assertEqual(synd.pages, [post2, post1])

            self.assertIsNone(post1.syndication)
            self.assertIsNone(post2.syndication)
            self.assertIsNone(widget.syndication)

            self.assertEqual(rss.pages, synd.pages)
            self.assertEqual(rss.title, "Syndication")

            self.assertEqual(atom.pages, synd.pages)
            self.assertEqual(atom.title, "Syndication")

    def test_complex(self):
        files = dict(BASE_FILES)
        files[
            "blog.md"
        ] = """---
date: 2016-04-16 10:23:00+02:00
pages: blog/*2.*
syndication:
  add_to:
    path: blog/*1.*
---

# Title

text
"""
        with self.site(files) as mocksite:
            blog, post1, post2, widget, rss, atom = mocksite.page(
                "blog",
                "blog/post1",
                "blog/post2",
                "blog/widget",
                "blog/index.rss",
                "blog/index.atom",
            )

            synd = blog.syndication
            self.assertEqual(synd.pages, [post2])

            self.assertIsNone(post1.syndication)
            self.assertEqual(
                post1.related,
                {
                    "rss_feed": rss,
                    "atom_feed": atom,
                },
            )
            self.assertIsNone(post2.syndication)
            self.assertIsNone(widget.syndication)
            self.assertEqual(rss.pages, synd.pages)
            self.assertEqual(atom.pages, synd.pages)



from __future__ import annotations

import heapq
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..page import Page


def arrange(pages: list[Page], sort: str, limit: int | None = None) -> list[Page]:
    """
    Sort the pages by ``sort`` and take the first ``limit`` ones
    """
    from ..page_filter import sort_args

    sort_meta, reverse, key = sort_args(sort)
    if key is None:
        if limit is None:
            return pages
        else:
            return pages[:limit]
    else:
        if limit is None:
            return sorted(pages, key=key, reverse=reverse)
        elif limit == 1:
            if reverse:
                return [max(pages, key=key)]
            else:
                return [min(pages, key=key)]
        elif len(pages) > 10 and limit < len(pages) / 3:
            if reverse:
                return heapq.nlargest(limit, pages, key=key)
            else:
                return heapq.nsmallest(limit, pages, key=key)
        else:
            return sorted(pages, key=key, reverse=reverse)[:limit]



from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING, cast

from staticsite.cmd.site import FeatureCommand

if TYPE_CHECKING:
    from staticsite.features.data import DataPages
    from staticsite.features.links import LinksPage

log = logging.getLogger("links")


class LinkLint(FeatureCommand):
    "run consistency checks on links data"

    NAME = "link_lint"

    def run(self) -> None:
        # links = self.site.features["links"]
        seen = defaultdict(list)
        data = cast(DataPages, self.site.features["data"])
        for lpage in data.by_type.get("links", ()):
            page = cast(LinksPage, lpage)
            if not page.links:
                continue
            for link in page.links:
                seen[link.url].append(page)

        for url, pages in seen.items():
            if len(pages) == 1:
                continue
            print(f"{url}: link appears in multiple pages:")
            for page in pages:
                print(f" - {page.src.relpath}")



from __future__ import annotations

import datetime
import logging
import os
import re
import zoneinfo
from collections.abc import Generator, Iterable
from functools import cached_property
from typing import TYPE_CHECKING, Any

import dateutil.parser
import pytz

from . import fields, fstree
from .cache import Caches, DisabledCaches
from .file import File
from .settings import Settings
from .utils import timings

if TYPE_CHECKING:
    from .archetypes import Archetypes
    from .node import Node
    from .page import Page, SourcePage
    from .source_node import RootNode, SourceNode
    from .theme import Theme

log = logging.getLogger("site")


class Path(tuple[str]):
    """
    Path in the site, split into components
    """

    re_pathsep = re.compile(re.escape(os.sep) + "+")

    # TODO: this is not needed, but mypy is currently unable to understand that
    # Path can be False
    def __bool__(self) -> bool:
        return len(self) > 0

    @property
    def head(self) -> str:
        """
        Return the first element in the path
        """
        return self[0]

    @property
    def tail(self) -> Path:
        """
        Return the Path after the first element
        """
        # TODO: check if it's worth making this a cached_property
        return Path(self[1:])

    @property
    def dir(self) -> Path:
        """
        Return the path with all components except last
        """
        return Path(self[:-1])

    @property
    def name(self) -> str:
        """
        Return the last component of the path
        """
        return self[-1]

    @classmethod
    def from_string(cls, path: str) -> Path:
        """
        Split a string into a path
        """
        return cls(cls.re_pathsep.split(path.strip(os.sep)))


class SiteElement(fields.FieldContainer):
    """
    Common fields for site elements
    """

    site_name = fields.Str["SiteElement"](
        inherited=True,
        doc="""
        Name of the site. If missing, it defaults to the title of the toplevel index
        page. If missing, it defaults to the name of the content directory.
    """,
    )

    site_url = fields.Str["SiteElement"](
        inherited=True,
        doc="""
        Base URL for the site, used to generate an absolute URL to the page.
    """,
    )

    author = fields.Str["SiteElement"](
        inherited=True,
        doc="""
        A string with the name of the author for this page.

        SITE_AUTHOR is used as a default if found in settings.

        If not found, it defaults to the current user's name.
    """,
    )

    template_copyright = fields.Template["SiteElement"](
        inherited=True,
        doc="""
        jinja2 template to use to generate `copyright` when it is not explicitly set.

        The template context will have `page` available, with the current page. The
        result of the template will not be further escaped, so you can use HTML markup
        in it.

        If missing, defaults to `"© {{meta.date.year}} {{meta.author}}"`
    """,
    )

    template_title = fields.Template["SiteElement"](
        inherited=True,
        doc="""
        jinja2 template to use to generate `title` when it is not explicitly set.

        The template context will have `page` available, with the current page.
        The result of the template will not be further escaped, so you can use
        HTML markup in it.
    """,
    )

    template_description = fields.Template["SiteElement"](
        inherited=True,
        doc="""
        jinja2 template to use to generate `description` when it is not
        explicitly set.

        The template context will have `page` available, with the current page.
        The result of the template will not be further escaped, so you can use
        HTML markup in it.
    """,
    )

    asset = fields.Bool["SiteElement"](
        inherited=True,
        doc="""
        If set to True for a file (for example, by a `file:` pattern in a directory
        index), the file is loaded as a static asset, regardless of whether a feature
        would load it.

        If set to True in a directory index, the directory and all its subdirectories
        are loaded as static assets, without the interventions of features.
    """,
    )

    def lookup_page(self, path: Path) -> Page | None:
        """
        Find a page by given a path starting from this site element

        The final path component can match:
        * a node name
        * the basename of a page's src_relpath
        * the rendered file name
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}.lookup_page() not implemented"
        )

    def resolve_path(self, target: str | Page, static: bool = False) -> Page:
        """
        Return a Page from the site, given a source or site path relative to
        this page.

        The path is resolved relative to this node, and if not found, relative
        to the parent node, and so on until the top.
        """
        from .page import Page, PageNotFoundError

        if isinstance(target, Page):
            return target

        path = Path.from_string(target)

        root: SiteElement
        if target.startswith("/"):
            if static:
                root = self.site.static_root
            else:
                root = self.site.root
        else:
            root = self

        page = root.lookup_page(path)

        if page is None:
            raise PageNotFoundError(f"cannot resolve {target!r} relative to {root!r}")
        else:
            return page


class Site:
    """
    A staticsite site.

    This class tracks all resources associated with a site.
    """

    # Identifiers for steps of the site loading
    LOAD_STEP_INITIAL = 0
    LOAD_STEP_FEATURES = 1
    LOAD_STEP_THEME = 2
    LOAD_STEP_DIRS = 3
    LOAD_STEP_CONTENTS = 4
    LOAD_STEP_ORGANIZE = 5
    LOAD_STEP_GENERATE = 6
    LOAD_STEP_CROSSREFERENCE = 7
    LOAD_STEP_ALL = LOAD_STEP_CROSSREFERENCE

    def __init__(
        self,
        settings: Settings | None = None,
        generation_time: datetime.datetime | None = None,
    ):
        from .feature import Features

        # Site settings
        if settings is None:
            settings = Settings()
        self.settings: Settings = settings

        # Fill in settings left empty that have meaningful defaults
        if self.settings.PROJECT_ROOT is None:
            self.settings.PROJECT_ROOT = os.getcwd()
        if self.settings.CONTENT is None:
            self.settings.CONTENT = "."

        # Site time zone
        self.timezone: datetime.tzinfo
        if settings.TIMEZONE is None:
            from dateutil.tz import tzlocal

            self.timezone = tzlocal()
        else:
            self.timezone = zoneinfo.ZoneInfo(settings.TIMEZONE)

        # Current datetime
        self.generation_time: datetime.datetime
        if generation_time is not None:
            self.generation_time = generation_time.astimezone(self.timezone)
        else:
            self.generation_time = pytz.utc.localize(
                datetime.datetime.utcnow()
            ).astimezone(self.timezone)

        # Filesystem trees scanned by the site
        self.fstrees: dict[str, fstree.Tree] = {}

        # Root node of the site
        self.root: RootNode

        # Root node for the static contents of the site
        self.static_root: SourceNode

        # Last load step performed
        self.last_load_step = self.LOAD_STEP_INITIAL

        # Set to True when feature constructors have been called
        self.stage_features_constructed = False

        # Theme used to render pages
        self._theme: Theme | None = None

        # Feature implementation registry
        self.features = Features(self)

        # Build cache repository
        self.caches: Caches | DisabledCaches
        if self.settings.CACHE_REBUILDS:
            if os.access(self.settings.PROJECT_ROOT, os.W_OK):
                self.caches = Caches(
                    os.path.join(self.settings.PROJECT_ROOT, ".staticsite-cache")
                )
            else:
                log.warning(
                    "%s: directory not writable: disabling caching",
                    self.settings.PROJECT_ROOT,
                )
                self.caches = DisabledCaches()
        else:
            self.caches = DisabledCaches()

        # Content root for the website
        self.content_root = os.path.normpath(
            os.path.join(self.settings.PROJECT_ROOT, self.settings.CONTENT)
        )

        # Cache with last build information
        self.build_cache = self.caches.get("build")

        # Source page footprints from a previous build
        self.previous_source_footprints: dict[str, dict[str, Any]]

        # Pages for which we should call Page.crossreference() at the beginning of the crossreference stage
        self.pages_to_crossreference: set[Page] = set()

    @cached_property
    def theme(self) -> Theme:
        """
        Return the Theme object for this site
        """
        if self._theme is None:
            raise RuntimeError("Site.theme called before the theme was loaded")
        return self._theme

    def clear_footprints(self) -> None:
        """
        Clear the previous-build page footprints.

        This is used when something failed during a build, and the build
        directory is left in an inconsistent state
        """
        self.build_cache.put("footprints", {})

    def save_footprints(self) -> None:
        """
        Save the current set of page footprints as reference for a future build

        This is used after a build completed successfully, to allow incremental
        future builds
        """
        footprints = {
            page.src.relpath: page.footprint for page in self._iter_source_pages()
        }
        self.build_cache.put("footprints", footprints)
        for feature in self.features.ordered():
            self.build_cache.put(f"footprint_{feature.name}", feature.get_footprint())

    def deleted_source_pages(self) -> Iterable[str]:
        """
        Return a sequence with the source relpaths of pages deleted since the
        last run
        """
        # Entries are popped from previous_source_footprints while pages are
        # loaded, so if any remain it means that those pages have been deleted
        return self.previous_source_footprints.keys()

    def iter_pages(
        self, static: bool = True, source_only: bool = False
    ) -> Generator[Page, None, None]:
        """
        Iterate all pages in the site
        """
        prune: tuple[Node, ...]
        if static:
            prune = ()
        elif self.root != self.static_root:
            prune = (self.static_root,)
        else:
            prune = ()
        yield from self.root.iter_pages(prune=prune, source_only=source_only)

    def _iter_source_pages(self) -> Generator[SourcePage, None, None]:
        """
        Iterate all source pages in the site.

        This is the same as iter_pages(source_only=True), but has the right generator type
        """
        # Ignore typing, as casting the generator currently requires importing
        # SourcePage at runtime
        yield from self.root.iter_pages(source_only=True)  # type: ignore

    @cached_property
    def archetypes(self) -> Archetypes:
        """
        Archetypes defined in the site
        """
        from .archetypes import Archetypes

        if self.settings.PROJECT_ROOT is None:
            raise RuntimeError("PROJECT_ROOT is not set and cannot be inferred")
        return Archetypes(self, os.path.join(self.settings.PROJECT_ROOT, "archetypes"))

    def load_features(self) -> None:
        """
        Load default features
        """
        self.features.load_default_features()
        # We can now load source files footprint, before theme loads assets
        previous_source_footprints = self.build_cache.get("footprints")
        if previous_source_footprints is None:
            self.previous_source_footprints = {}
        else:
            self.previous_source_footprints = previous_source_footprints

    def load_theme(self) -> None:
        """
        Load a theme from the given directory.

        This needs to be called once (and only once) before analyze() is
        called.
        """
        from .theme import Theme

        if self._theme is not None:
            raise RuntimeError(
                f"load_theme called while theme {self._theme.name} was already loaded"
            )

        if isinstance(self.settings.THEME, str):
            self._theme = Theme.create(self, self.settings.THEME)
        else:
            # Pick the first valid theme directory
            candidate_themes = self.settings.THEME
            if isinstance(candidate_themes, str):
                candidate_themes = (candidate_themes,)
            self._theme = Theme.create_legacy(self, candidate_themes)

        self._theme.load()

        # We now have the final feature list, we can load old feature footprints
        for feature in self.features.ordered():
            footprint = self.build_cache.get(f"footprint_{feature.name}")
            feature.set_previous_footprint(footprint if footprint is not None else {})

    def _settings_to_meta(self) -> None:
        """
        Build directory metadata based on site settings
        """
        self.root.template_copyright = "© {{page.meta.date.year}} {{page.meta.author}}"
        if self.settings.SITE_URL:
            self.root.site_url = self.settings.SITE_URL
        if self.settings.SITE_ROOT:
            self.root.site_path = os.path.normpath(
                os.path.join("/", self.settings.SITE_ROOT)
            )
        else:
            self.root.site_path = "/"
        if self.settings.SITE_NAME:
            self.root.site_name = self.settings.SITE_NAME
        if self.settings.SITE_AUTHOR:
            self.root.author = self.settings.SITE_AUTHOR
        else:
            import getpass
            import pwd

            user = getpass.getuser()
            pw = pwd.getpwnam(user)
            self.root.author = pw.pw_gecos.split(",")[0]

        # log.debug("Initial settings: %r", res)

    def scan_content(self) -> None:
        """
        Scan content root directories, building metadata for the directories in
        the site tree
        """
        from .source_node import RootNode

        if not self.stage_features_constructed:
            log.warning("scan_content called before site features have been loaded")

        if not os.path.exists(self.content_root):
            log.info("%s: content tree does not exist", self.content_root)
            return

        src = File.with_stat("", os.path.abspath(self.content_root))

        # Create root node
        self.root = self.features.get_node_class(RootNode)(self, src=src)
        self._settings_to_meta()

        # Create static root node
        self.static_root = self.root.static_root(
            Path.from_string(self.settings.STATIC_PATH)
        )

        # Scan the main content filesystem
        tree = self.scan_content_tree(src)

        if self.root.site_name is None:
            # If we still have no idea of site names, use the root directory's name
            self.root.site_name = os.path.basename(tree.src.abspath)

        # Scan asset trees from themes
        self.theme.scan_assets()

    def scan_content_tree(self, src: File) -> fstree.Tree:
        """
        Scan the contents of the given directory, adding it to self.fstrees
        """
        if src.abspath in self.fstrees:
            # TODO: just return?
            raise RuntimeError(f"{src.abspath} scanned twice")
        tree = fstree.RootPageTree(site=self, src=src, node=self.root)
        with tree.open_tree():
            tree.scan()
        self.fstrees[src.abspath] = tree
        return tree

    def scan_asset_tree(self, src: File, node: SourceNode) -> fstree.Tree:
        """
        Scan the contents of the given directory, adding it to self.fstrees
        """
        if src.abspath in self.fstrees:
            # TODO: just return?
            raise RuntimeError(f"{src.abspath} scanned twice")
        tree = fstree.AssetTree(site=self, src=src, node=node)
        with tree.open_tree():
            tree.scan()
        self.fstrees[src.abspath] = tree
        return tree

    def load_content(self) -> None:
        """
        Load site page and assets from scanned content roots.
        """
        # Turn scanned filesytem information into site structure
        for abspath, tree in self.fstrees.items():
            with tree.open_tree():
                tree.populate_node()

    def load(self, until: int = LOAD_STEP_ALL) -> None:
        """
        Load all site components
        """
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_FEATURES:
            with timings("Loaded default features in %fs"):
                self.load_features()
            self.last_load_step = self.LOAD_STEP_FEATURES
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_THEME:
            with timings("Loaded theme in %fs"):
                self.load_theme()
            self.last_load_step = self.LOAD_STEP_THEME
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_DIRS:
            with timings("Scanned contents in %fs"):
                self.scan_content()
            self.last_load_step = self.LOAD_STEP_DIRS
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_CONTENTS:
            with timings("Loaded contents in %fs"):
                self.load_content()
            self.last_load_step = self.LOAD_STEP_CONTENTS
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_ORGANIZE:
            with timings("Organized contents in %fs"):
                self._organize()
            self.last_load_step = self.LOAD_STEP_ORGANIZE
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_GENERATE:
            with timings("Generated new contents in %fs"):
                self._generate()
            self.last_load_step = self.LOAD_STEP_GENERATE
        if until <= self.last_load_step:
            return

        if self.last_load_step < self.LOAD_STEP_CROSSREFERENCE:
            with timings("Cross-referenced contents in %fs"):
                self._crossreference()
            self.last_load_step = self.LOAD_STEP_CROSSREFERENCE
        if until <= self.last_load_step:
            return

    def _organize(self) -> None:
        """
        Call features to organize the pages that have just been loaded from
        sources
        """
        # Call organize hook on features
        for feature in self.features.ordered():
            feature.organize()

    def _generate(self) -> None:
        """
        Call features to generate new pages
        """
        # Call generate hook on features
        for feature in self.features.ordered():
            feature.generate()

    def _crossreference(self) -> None:
        """
        Call features to cross-reference pages after the site structure has
        been finalized
        """
        # Call crossreference hook on selected pages
        for page in self.pages_to_crossreference:
            page.crossreference()

        # Call crossreference hook on features
        for feature in self.features.ordered():
            feature.crossreference()

    def slugify(self, text: str) -> str:
        """
        Return the slug version of an arbitrary string, that can be used as an
        url component or file name.
        """
        from slugify import slugify

        return slugify(text)

    def localized_timestamp(self, ts: float) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(ts, tz=pytz.utc).astimezone(
            self.timezone
        )

    # TODO: support partial dates?
    re_isodate = re.compile(r"^(\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2})(Z|\+\d{2}.*)$")

    def clean_date(self, date: str | datetime.datetime) -> datetime.datetime:
        """
        Return an aware datetime from a potential date value.

        Parse a date string, or make sure a datetime value is aware, or replace
        None with the current generation time
        """
        clean_date: datetime.datetime
        # Make sure we have a datetime
        if not isinstance(date, datetime.datetime):
            mo = self.re_isodate.match(date)
            if mo:
                if mo.group(2) == "Z":
                    clean_date = datetime.datetime.fromisoformat(mo.group(1)).replace(
                        tzinfo=pytz.utc
                    )
                else:
                    clean_date = datetime.datetime.fromisoformat(date)
            else:
                # TODO: log a fallback on dateutil?
                try:
                    clean_date = dateutil.parser.parse(date)
                except ValueError as e:
                    log.warning("cannot parse datetime %s: %s", date, e)
                    return self.generation_time
        else:
            clean_date = date

        # Make sure the datetime is aware
        if clean_date.tzinfo is None:
            if hasattr(self.timezone, "localize"):
                clean_date = self.timezone.localize(clean_date)
            else:
                clean_date = clean_date.replace(tzinfo=self.timezone)

        return clean_date
